import time
import pytest
from ymmsl import CheckpointRangeRule, CheckpointAtRule, Checkpoints

from libmuscle.checkpoint_triggers import (
    CombinedCheckpointTriggers, AtCheckpointTrigger, RangeCheckpointTrigger,
    TriggerManager)


def test_at_checkpoint_trigger():
    trigger = AtCheckpointTrigger([CheckpointAtRule([1.0, 3.0, 4.0, 4.5, 9.0])])

    assert trigger.next_checkpoint(0) == 1
    assert trigger.previous_checkpoint(0) is None

    assert trigger.next_checkpoint(1) == 3
    assert trigger.previous_checkpoint(1) == 1

    eps = 1e-16
    assert trigger.next_checkpoint(1 - eps) == 1
    assert trigger.previous_checkpoint(1 - eps) is None

    assert trigger.next_checkpoint(3.9) == 4
    assert trigger.previous_checkpoint(3.9) == 3

    assert trigger.next_checkpoint(4.1) == 4.5
    assert trigger.previous_checkpoint(4.1) == 4

    assert trigger.next_checkpoint(5) == 9
    assert trigger.previous_checkpoint(5) == 4.5

    assert trigger.next_checkpoint(9) is None
    assert trigger.previous_checkpoint(9) == 9

    assert trigger.next_checkpoint(11) is None
    assert trigger.previous_checkpoint(11) == 9


def test_range_checkpoint_trigger():
    range = CheckpointRangeRule(start=0.0, stop=20.0, every=1.2)
    trigger = RangeCheckpointTrigger(range)

    assert trigger.next_checkpoint(-1) == 0
    assert trigger.previous_checkpoint(-1) is None

    assert trigger.next_checkpoint(0) == pytest.approx(1.2)
    assert trigger.previous_checkpoint(0) == 0

    assert trigger.next_checkpoint(8) == pytest.approx(8.4)
    assert trigger.previous_checkpoint(8) == pytest.approx(7.2)

    assert trigger.next_checkpoint(18.2) == pytest.approx(19.2)
    assert trigger.previous_checkpoint(18.2) == pytest.approx(18)

    assert trigger.next_checkpoint(20) is None
    assert trigger.previous_checkpoint(20) == pytest.approx(19.2)


def test_range_checkpoint_trigger_default_stop():
    range = CheckpointRangeRule(start=1.0, every=1.2)
    trigger = RangeCheckpointTrigger(range)

    assert trigger.next_checkpoint(-1.) == 1
    assert trigger.previous_checkpoint(-1.) is None

    assert trigger.next_checkpoint(148148.) == pytest.approx(148148.2)
    assert trigger.previous_checkpoint(148148.) == pytest.approx(148147)

    assert trigger.next_checkpoint(148148148.) == pytest.approx(148148149)
    assert trigger.previous_checkpoint(148148148.) == pytest.approx(148148147.8)


def test_range_checkpoint_trigger_default_start():
    range = CheckpointRangeRule(every=1.2, stop=10.0)
    trigger = RangeCheckpointTrigger(range)

    assert trigger.next_checkpoint(10) is None
    assert trigger.previous_checkpoint(10) == pytest.approx(9.6)

    assert trigger.next_checkpoint(0.0) == pytest.approx(1.2)
    assert trigger.previous_checkpoint(0.0) == pytest.approx(0.0)

    assert trigger.next_checkpoint(-148148.) == pytest.approx(-148147.2)
    assert trigger.previous_checkpoint(-148148.) == pytest.approx(-148148.4)


def test_combined_checkpoint_trigger_every_at():
    rules = [CheckpointRangeRule(every=10.0), CheckpointAtRule([3.0, 7.0, 13.0, 17.0])]
    trigger = CombinedCheckpointTriggers(rules)

    assert trigger.next_checkpoint(-11.) == pytest.approx(-10)
    assert trigger.previous_checkpoint(-11) == pytest.approx(-20)

    assert trigger.next_checkpoint(0.) == pytest.approx(3)
    assert trigger.previous_checkpoint(0.) == pytest.approx(0)

    assert trigger.next_checkpoint(8.3) == pytest.approx(10)
    assert trigger.previous_checkpoint(8.3) == pytest.approx(7)

    assert trigger.next_checkpoint(14.2) == pytest.approx(17)
    assert trigger.previous_checkpoint(14.2) == pytest.approx(13)

    assert trigger.next_checkpoint(25.2) == pytest.approx(30)
    assert trigger.previous_checkpoint(25.2) == pytest.approx(20)


def test_combined_checkpoint_trigger_at_ranges():
    rules = [CheckpointAtRule([3.0, 7.0, 13.0, 17.0]),
             CheckpointRangeRule(start=0.0, every=5.0, stop=20.0),
             CheckpointRangeRule(start=20.0, every=20.0, stop=100.0)]
    trigger = CombinedCheckpointTriggers(rules)

    assert trigger.next_checkpoint(-11.) == pytest.approx(0)
    assert trigger.previous_checkpoint(-11) is None

    assert trigger.next_checkpoint(0.) == pytest.approx(3)
    assert trigger.previous_checkpoint(0.) == pytest.approx(0)

    assert trigger.next_checkpoint(8.3) == pytest.approx(10)
    assert trigger.previous_checkpoint(8.3) == pytest.approx(7)

    assert trigger.next_checkpoint(14.2) == pytest.approx(15)
    assert trigger.previous_checkpoint(14.2) == pytest.approx(13)

    assert trigger.next_checkpoint(19.3) == pytest.approx(20)
    assert trigger.previous_checkpoint(19.3) == pytest.approx(17)

    assert trigger.next_checkpoint(25.2) == pytest.approx(40)
    assert trigger.previous_checkpoint(25.2) == pytest.approx(20)

    assert trigger.next_checkpoint(95.2) == pytest.approx(100)
    assert trigger.previous_checkpoint(95.2) == pytest.approx(80)

    assert trigger.next_checkpoint(125.2) is None
    assert trigger.previous_checkpoint(125.2) == pytest.approx(100)


def test_trigger_manager_reference_time():
    monotonic_start = time.monotonic()
    ref_elapsed = 15.0
    trigger_manager = TriggerManager()
    trigger_manager.set_checkpoint_info(ref_elapsed, Checkpoints(at_end=True))
    elapsed_walltime = trigger_manager.elapsed_walltime()
    duration = time.monotonic() - monotonic_start
    assert ref_elapsed < elapsed_walltime < (ref_elapsed + duration)


def test_trigger_manager():
    ref_elapsed = 0.0
    trigger_manager = TriggerManager()
    trigger_manager.set_checkpoint_info(ref_elapsed, Checkpoints(
            at_end=True,
            wallclock_time=[CheckpointAtRule([1e-12])],
            simulation_time=[CheckpointAtRule([1.0, 3.0, 5.0])]))

    assert trigger_manager.should_save_snapshot(0.1)
    triggers = trigger_manager.get_triggers()
    assert len(triggers) == 1
    assert "wallclock_time" in triggers[0]
    trigger_manager.update_checkpoints(0.1)

    assert not trigger_manager.should_save_snapshot(0.99)

    assert trigger_manager.should_save_snapshot(3.2)
    triggers = trigger_manager.get_triggers()
    assert len(triggers) == 1
    assert "simulation_time" in triggers[0]
    trigger_manager.update_checkpoints(3.2)

    assert trigger_manager.should_save_final_snapshot(True, 7.0)
    assert len(trigger_manager.get_triggers()) > 0
    trigger_manager.update_checkpoints(7.0)

    assert not trigger_manager.should_save_snapshot(7.1)

    assert trigger_manager.should_save_final_snapshot(False, None)
    trigger_manager.update_checkpoints(7.1)


def test_no_checkpointing() -> None:
    trigger_manager = TriggerManager()
    trigger_manager.set_checkpoint_info(0.0, Checkpoints())

    assert not trigger_manager.should_save_snapshot(1)
    assert not trigger_manager.should_save_snapshot(5000)
    assert not trigger_manager.should_save_final_snapshot(False, None)



from itertools import product
import logging
import os
from parsimonious import Grammar, NodeVisitor
from parsimonious.nodes import Node
import subprocess
from typing import Any, cast, List, Optional, Sequence, Tuple


_logger = logging.getLogger(__name__)


_node_range_expression_grammar = Grammar(
        """
        nre = nre_parts ("," nre_parts)*
        nre_parts = nre_part+
        nre_part = identifier ("[" index_set "]")?
        index_set = index_range ("," index_range)*
        index_range = integer ("-" integer)?
        identifier = ~"[a-z 0-9 _-]+"i
        integer = padded_int / int
        int = ~"[0-9]+"
        padded_int = ~"0[0-9]+"
        """
        )


class NREVisitor(NodeVisitor):
    """Processes a parsed NRE and produces a list of nodes.

    Node range expressions are used by SLURM to describe collections of nodes. See
    parse_slurm_nodelist() below.
    """
    def visit_nre(
            self, node: Node,
            visited_children: Tuple[List[str], Sequence[Tuple[Any, List[str]]]]
            ) -> List[str]:
        """Return a list of nodes corresponding to the NRE."""
        nodes = visited_children[0].copy()
        for _, more_nodes in visited_children[1]:
            nodes.extend(more_nodes)
        return nodes

    def visit_nre_parts(
            self, node: Node, visited_children: Sequence[Tuple[str, List[str]]]
            ) -> List[str]:
        """Return a list of node ids for the part."""
        fmt = ''.join([c[0] + '{}' for c in visited_children])
        index_lists = [c[1] for c in visited_children]
        return [fmt.format(*idxs) for idxs in product(*index_lists)]

    def visit_nre_part(
            self, node: Node, visited_children: Tuple[
                str, Sequence[Tuple[Any, List[str], Any]]]
            ) -> Tuple[str, List[str]]:
        """Return the identifier part and a list of indexes for the set."""
        identifier = visited_children[0]
        if not visited_children[1]:
            index_set = ['']
        else:
            index_set = visited_children[1][0][1]
        return identifier, index_set

    def visit_index_set(
            self, node: Node,
            visited_children: Tuple[List[str], Sequence[Tuple[Any, List[str]]]]
            ) -> List[str]:
        """Return a list of indexes corresponding to the set."""
        indexes = visited_children[0].copy()
        for _, more_indexes in visited_children[1]:
            indexes.extend(more_indexes)
        return indexes

    def visit_index_range(
            self, node: Node,
            visited_children: Tuple[
                Tuple[int, int],
                Sequence[
                    Tuple[Any, Tuple[int, int]]
                    ]]
            ) -> List[str]:
        """Return a list of indexes corresponding to the range."""

        def format_str(width: int) -> str:
            if width == -1:
                return '{}'
            return f'{{:0{width}}}'

        start_value, width = visited_children[0]
        if visited_children[1]:
            end_value, _ = visited_children[1][0][1]
            fmt = format_str(width)
            return [fmt.format(i) for i in range(start_value, end_value + 1)]

        fmt = format_str(width)
        return [fmt.format(start_value)]

    def visit_identifier(self, node: Node, _: Sequence[Any]) -> str:
        return node.text

    def visit_integer(
            self, node: Node, visited_children: Sequence[Tuple[int, int]]
            ) -> Tuple[int, int]:
        """Returns the value of the int, and a field width or -1."""
        return visited_children[0]

    def visit_int(self, node: Node, _: Sequence[Any]) -> Tuple[int, int]:
        """Returns the value and a field width of -1."""
        return int(node.text), -1

    def visit_padded_int(self, node: Node, _: Sequence[Any]) -> Tuple[int, int]:
        """Returns the value of the int and the field width."""
        return int(node.text), len(node.text)

    def generic_visit(
            self, node: Node, visited_children: Sequence[Any]) -> Sequence[Any]:
        return visited_children


_nre_visitor = NREVisitor()


def parse_slurm_nodelist(s: str) -> List[str]:
    """Parse a SLURM node range expression and produce node names.

    Exactly what the syntax is for a "node range expression" isn't entirely
    clear. Some examples are given throughout the documentation:

    linux[00-17]
    lx[10-20]
    tux[2,1-2]
    tux[1-2,2]
    tux[1-3]
    linux[0-64,128]
    alpha,beta,gamma
    lx[15,18,32-33]
    linux[0000-1023]
    rack[0-63]_blade[0-41]

    unit[0-31]rack is invalid

    If a range uses leading zeros, then so should the generated indexes.
    See _node_range_expression_grammar above for my best guess at the
    correct grammar.

    This function takes a string containing an NRE and returns the
    corresponding list of node names.
    """
    ast = _node_range_expression_grammar.parse(s)
    return cast(List[str], _nre_visitor.visit(ast))


_nodes_cores_expression_grammar = Grammar(
        """
        nce = nce_run ("," nce_run)*
        nce_run = int ("(" run_length ")")?
        run_length = "x" int
        int = ~"[0-9]+"
        """
        )


class NCEVisitor(NodeVisitor):
    """Processes a parsed NCE and produces a list of cpu counts per node.

    Nodes cores expressions are used by SLURM to describe cores on a collection of
    nodes.  See parse_slurm_nodes_cores() below.
    """
    def visit_nce(
            self, node: Node,
            visited_children: Tuple[List[int], Sequence[Tuple[Any, List[int]]]]
            ) -> List[int]:
        """Return a list of nodes corresponding to the NRE."""
        nodes_cores = visited_children[0].copy()
        for _, more_nodes_cores in visited_children[1]:
            nodes_cores.extend(more_nodes_cores)
        return nodes_cores

    def visit_nce_run(
            self, node: Node,
            visited_children: Tuple[int, Sequence[Tuple[Any, int, Any]]]
            ) -> List[int]:
        """Return a list of core counts produced by this run."""
        num_cores = visited_children[0]
        result = [num_cores]

        if visited_children[1]:
            result *= visited_children[1][0][1]

        return result

    def visit_run_length(
            self, node: Node, visited_children: Tuple[str, int]) -> int:
        """Return the number of repetitions."""
        return visited_children[1]

    def visit_int(self, node: Node, _: Sequence[Any]) -> int:
        """Returns the value as an int"""
        return int(node.text)

    def generic_visit(
            self, node: Node, visited_children: Sequence[Any]) -> Sequence[Any]:
        return visited_children


_nce_visitor = NCEVisitor()


def parse_slurm_nodes_cores(s: str) -> List[int]:
    """Parse a SLURM nodes cores expression and produce node names.

    The sbatch documentation page describes the format under
    SLURM_JOB_CPUS_PER_NODE as CPU_count[(xnumber_of_nodes)][,CPU_count
    [(xnumber_of_nodes)] ...]. and gives the example of '72(x2),36' describing a set of
    three nodes, the first two with 72 cores and the third with 36.

    See _nodes_cores_expression_grammar above for the corresponding grammar.

    This function takes a string containing an NCE and returns the corresponding list of
    node names.
    """
    ast = _nodes_cores_expression_grammar.parse(s)
    return cast(List[int], _nce_visitor.visit(ast))


class SlurmQuirks:
    """Collects features of the present SLURM."""
    overlap: bool
    """True iff --overlap must be specified for srun."""
    cpu_bind: str
    """CPU binding argument, --cpu-bind or --cpu_bind."""


class SlurmInfo:
    """Detects and holds information about the present SLURM scheduler."""
    def __init__(self) -> None:
        if self.in_slurm_allocation():
            self.version = self._slurm_version()
            self.quirks = SlurmQuirks()

            self.quirks.overlap = self.version > (20, 2)
            self.quirks.cpu_bind = (
                    '--cpu-bind' if self.version > (17, 2) else '--cpu_bind')

    def in_slurm_allocation(self) -> bool:
        """Check whether we're in a SLURM allocation.

        Returns true iff SLURM was detected.
        """
        return 'SLURM_JOB_ID' in os.environ

    def get_nodes(self) -> List[str]:
        """Get a list of node names from SLURM_JOB_NODELIST.

        This inspects SLURM_JOB_NODELIST or SLURM_NODELIST and returns an
        expanded list of node names.

        If SLURM_JOB_NODELIST is "node[020-023]" then this returns
        ["node020", "node021", "node022", "node023"].
        """
        nodelist = os.environ.get('SLURM_JOB_NODELIST')
        if not nodelist:
            nodelist = os.environ.get('SLURM_NODELIST')
        if not nodelist:
            raise RuntimeError('SLURM_(JOB_)NODELIST not set, are we running locally?')

        _logger.debug(f'SLURM node list: {nodelist}')

        return parse_slurm_nodelist(nodelist)

    def get_logical_cpus_per_node(self) -> List[int]:
        """Return the number of logical CPU cores per node.

        This returns a list with the number of cores of each node in the result of
        get_nodes(), which gets read from SLURM_JOB_CPUS_PER_NODE.
        """
        sjcpn = os.environ.get('SLURM_JOB_CPUS_PER_NODE')
        _logger.debug(f'SLURM_JOB_CPUS_PER_NODE: {sjcpn}')

        if sjcpn:
            return parse_slurm_nodes_cores(sjcpn)
        else:
            scon = os.environ.get('SLURM_CPUS_ON_NODE')
            _logger.debug(f'SLURM_CPUS_ON_NODE: {scon}')

            snn = os.environ.get('SLURM_JOB_NUM_NODES')
            if not snn:
                snn = os.environ.get('SLURM_NNODES')
            _logger.debug(f'SLURM num nodes: {snn}')

            if scon and snn:
                return [int(scon)] * int(snn)

        raise RuntimeError(
                'SLURM_JOB_CPUS_PER_NODE is not set in the environment, and also'
                ' SLURM_CPUS_ON_NODE is missing or neither SLURM_JOB_NUM_NODES nor'
                ' SLURM_NNODES is set. Please create an issue on GitHub with the output'
                ' of "sbatch --version" on this cluster.')

    def agent_launch_command(self, agent_cmd: List[str], nnodes: int) -> List[str]:
        """Return a command for launching one agent on each node.

        Args:
            agent_cmd: A command that will start the agent.
        """
        # TODO: On the latest Slurm, there's a special command for this that we should
        # use if we have that, --external-launcher. Poorly documented though, so will
        # require some experimentation.

        # On SLURM <= 23-02, the number of tasks is inherited by srun from sbatch rather
        # than calculated anew from --nodes and --ntasks-per-node, so we specify it
        # explicitly to avoid getting an agent per logical cpu rather than per node.
        srun_cmd = [
                'srun', f'--nodes={nnodes}', f'--ntasks={nnodes}',
                '--ntasks-per-node=1'
                ]

        if self.quirks.overlap:
            srun_cmd.append('--overlap')

        return srun_cmd + agent_cmd

    def _slurm_version(self) -> Tuple[int, int]:
        """Obtains current version of SLURM from srun -v.

        This returns only the first two numbers, hopefully there won't be any changes in
        behaviour within a release series.
        """
        proc = subprocess.run(
                ['srun', '--version'], check=True, capture_output=True, text=True,
                encoding='utf-8'
                )

        output = proc.stdout.strip().split()
        if len(output) < 2:
            raise RuntimeError(
                    f'Unexpected srun version output "{output}". MUSCLE3 does not know'
                    ' how to run on this version of SLURM. Please file an issue on'
                    ' GitHub.')

        version_str = output[1]
        version = version_str.split('.')
        if len(version) < 2:
            _logger.error(f'srun produced unexpected version {version_str}')
            raise RuntimeError(
                    f'Unexpected srun version output "{output}". MUSCLE3 does not know'
                    ' how to run on this version of SLURM. Please file an issue on'
                    ' GitHub.')
        return int(version[0]), int(version[1])


slurm_: Optional[SlurmInfo] = None


def slurm() -> SlurmInfo:
    """SlurmInfo singleton"""
    global slurm_

    if slurm_ is None:
        slurm_ = SlurmInfo()

    return slurm_



from pathlib import Path
from typing import Any, List, Optional, Tuple

import msgpack

from libmuscle.mcp.protocol import AgentCommandType, RequestType, ResponseType
from libmuscle.mcp.tcp_transport_client import TcpTransportClient
from libmuscle.native_instantiator.agent.agent_commands import (
        AgentCommand, StartCommand, CancelAllCommand, ShutdownCommand)
from libmuscle.planner.resources import OnNodeResources


class MAPClient:
    """The client for the MUSCLE Agent Protocol.

    This class connects to the AgentManager and communicates with it.
    """
    def __init__(self, node_name: str, location: str) -> None:
        """Create a MAPClient

        Args:
            node_name: Name (hostname) of the local node
            location: A connection string of the form hostname:port
        """
        self._node_name = node_name
        self._transport_client = TcpTransportClient(location)

    def close(self) -> None:
        """Close the connection

        This closes the connection. After this no other member functions can be called.
        """
        self._transport_client.close()

    def report_resources(self, resources: OnNodeResources) -> None:
        """Report local resources

        Args:
            resources: Description of the resources on this node
        """
        enc_cpu_resources = [[c.cid] + list(c.hwthreads) for c in resources.cpu_cores]
        request = [
                RequestType.REPORT_RESOURCES.value,
                resources.node_name, {'cpu': enc_cpu_resources}]
        self._call_agent_manager(request)

    def get_command(self) -> Optional[AgentCommand]:
        """Get a command from the agent manager.

        Returns:
            A command, or None if there are no commands pending.
        """
        request = [RequestType.GET_COMMAND.value, self._node_name]
        response = self._call_agent_manager(request)

        if response[0] == ResponseType.PENDING.value:
            return None
        else:
            command = msgpack.unpackb(response[1], raw=False)

        if command[0] == AgentCommandType.START.value:
            name = command[1]
            workdir = Path(command[2])
            args = command[3]
            env = command[4]
            stdout = Path(command[5])
            stderr = Path(command[6])

            return StartCommand(name, workdir, args, env, stdout, stderr)

        elif command[0] == AgentCommandType.CANCEL_ALL.value:
            return CancelAllCommand()

        elif command[0] == AgentCommandType.SHUTDOWN.value:
            return ShutdownCommand()

        raise Exception('Unknown AgentCommand')

    def report_result(self, names_exit_codes: List[Tuple[str, int]]) -> None:
        """Report results of finished processes.

        Args:
            names_exit_codes: A list of names and exit codes of finished processes.
        """
        request = [RequestType.REPORT_RESULT.value, names_exit_codes]
        self._call_agent_manager(request)

    def _call_agent_manager(self, request: Any) -> Any:
        """Call the manager and do en/decoding.

        Args:
            request: The request to encode and send

        Returns:
            The decoded response
        """
        encoded_request = msgpack.packb(request, use_bin_type=True)
        response, _ = self._transport_client.call(encoded_request)
        return msgpack.unpackb(response, raw=False)



from logging import WARNING
from typing import Any
from unittest.mock import Mock

import pytest
from ymmsl import Operator, Reference

from libmuscle.port_manager import PortManager
from libmuscle.mmsf_validator import MMSFValidator


class Contains:
    """Helper class to simplify tests using caplog.record_tuples"""
    def __init__(self, value: Any) -> None:
        self.value = value

    def __eq__(self, other: Any) -> bool:
        return self.value in other

    def __repr__(self) -> str:
        return f"Contains({self.value!r})"


@pytest.fixture
def mock_peer_info() -> Mock:
    # Create a mock PeerInfo indicating that all ports are connected
    peer_info = Mock()
    peer_info.is_connected.return_value = True
    peer_info.get_peer_ports.return_value = [Reference("test")]
    peer_info.get_peer_dims.return_value = []
    return peer_info


@pytest.fixture
def validator_simple(mock_peer_info) -> MMSFValidator:
    port_manager = PortManager([], {
            Operator.F_INIT: ["f_i"],
            Operator.O_I: ["o_i"],
            Operator.S: ["s"],
            Operator.O_F: ["o_f"]})
    port_manager.connect_ports(mock_peer_info)
    return MMSFValidator(port_manager)


@pytest.mark.parametrize("num_iterations", [0, 1, 2])
@pytest.mark.parametrize("num_reuse", [1, 5])
def test_simple_correct(num_iterations, num_reuse, validator_simple, caplog):
    for _ in range(num_reuse):
        validator_simple.reuse_instance()
        validator_simple.check_receive("f_i", None)
        for _ in range(num_iterations):
            validator_simple.check_send("o_i", None)
            validator_simple.check_receive("s", None)
        validator_simple.check_send("o_f", None)
    # Final reuse_instance()
    validator_simple.reuse_instance()
    assert caplog.record_tuples == []


def test_simple_skip_f_init(validator_simple, caplog):
    validator_simple.reuse_instance()
    validator_simple.check_send("o_i", None)
    assert caplog.record_tuples == [
        ("libmuscle.mmsf_validator", WARNING, Contains("Send on port 'o_i'"))]


def test_simple_skip_o_i(validator_simple, caplog):
    validator_simple.reuse_instance()
    validator_simple.check_receive("f_i", None)

    validator_simple.check_receive("f_i", None)
    assert caplog.record_tuples == [
        ("libmuscle.mmsf_validator", WARNING, Contains("Receive on port 'f_i'"))]

    caplog.clear()
    validator_simple.check_receive("s", None)
    assert caplog.record_tuples == [
        ("libmuscle.mmsf_validator", WARNING, Contains("Receive on port 's'"))]

    caplog.clear()
    validator_simple.reuse_instance()
    assert caplog.record_tuples == [
        ("libmuscle.mmsf_validator", WARNING, Contains("reuse_instance()"))]


def test_simple_skip_s(validator_simple, caplog):
    validator_simple.reuse_instance()
    validator_simple.check_receive("f_i", None)
    validator_simple.check_send("o_i", None)

    validator_simple.check_send("o_i", None)
    assert caplog.record_tuples == [
        ("libmuscle.mmsf_validator", WARNING, Contains("Send on port 'o_i'"))]

    caplog.clear()
    validator_simple.check_send("o_f", None)
    assert caplog.record_tuples == [
        ("libmuscle.mmsf_validator", WARNING, Contains("Send on port 'o_f'"))]


def test_simple_skip_o_f(validator_simple, caplog):
    validator_simple.reuse_instance()
    validator_simple.check_receive("f_i", None)
    validator_simple.check_send("o_i", None)
    validator_simple.check_receive("s", None)

    validator_simple.reuse_instance()
    assert caplog.record_tuples == [
        ("libmuscle.mmsf_validator", WARNING, Contains("reuse_instance()"))]


def test_simple_skip_reuse_instance(validator_simple, caplog):
    validator_simple.reuse_instance()
    validator_simple.check_receive("f_i", None)
    validator_simple.check_receive("o_f", None)

    validator_simple.check_receive("f_i", None)
    assert caplog.record_tuples == [
        ("libmuscle.mmsf_validator", WARNING, Contains("Receive on port 'f_i'"))]


def test_only_o_f(mock_peer_info, caplog):
    port_manager = PortManager([], {Operator.O_F: ["o_f"]})
    port_manager.connect_ports(mock_peer_info)
    validator = MMSFValidator(port_manager)

    for _ in range(5):
        validator.reuse_instance()
        validator.check_send("o_f", None)

    validator.check_send("o_f", None)
    assert caplog.record_tuples == [
        ("libmuscle.mmsf_validator", WARNING, Contains("Send on port 'o_f'"))]


def test_only_f_i(mock_peer_info, caplog):
    port_manager = PortManager([], {Operator.F_INIT: ["f_i"]})
    port_manager.connect_ports(mock_peer_info)
    validator = MMSFValidator(port_manager)

    for _ in range(5):
        validator.reuse_instance()
        validator.check_receive("f_i", None)

    validator.check_receive("f_i", None)
    assert caplog.record_tuples == [
        ("libmuscle.mmsf_validator", WARNING, Contains("Receive on port 'f_i'"))]


def test_micro(mock_peer_info, caplog):
    port_manager = PortManager([], {Operator.F_INIT: ["f_i"], Operator.O_F: ["o_f"]})
    port_manager.connect_ports(mock_peer_info)
    validator = MMSFValidator(port_manager)

    for _ in range(5):
        validator.reuse_instance()
        validator.check_receive("f_i", None)
        validator.check_send("o_f", None)
    validator.reuse_instance()
    validator.check_receive("f_i", None)

    validator.check_receive("f_i", None)
    assert caplog.record_tuples == [
        ("libmuscle.mmsf_validator", WARNING, Contains("Receive on port 'f_i'"))]

    caplog.clear()
    validator.reuse_instance()
    assert caplog.record_tuples == [
        ("libmuscle.mmsf_validator", WARNING, Contains("reuse_instance()"))]


def test_not_all_ports_used(mock_peer_info, caplog):
    port_manager = PortManager([], {
            Operator.F_INIT: ["f_i1", "f_i2"], Operator.O_F: ["o_f"]})
    port_manager.connect_ports(mock_peer_info)
    validator = MMSFValidator(port_manager)

    validator.reuse_instance()
    validator.check_receive("f_i1", None)

    validator.check_send("o_f", None)
    assert caplog.record_tuples == [
        ("libmuscle.mmsf_validator", WARNING, Contains("Send on port 'o_f'"))]



import logging

from libmuscle.logging import LogLevel, LogMessage, Timestamp
from libmuscle.mmp_client import MMPClient


class MuscleManagerHandler(logging.Handler):
    """Standard Python log handler for the manager.

    A MuscleManagerHandler is a standard Python log handler, which can
    be attached to a logger, and forwards log messages to the Muscle
    Manager for central logging.
    """
    def __init__(self, instance_id: str, level: int, mmp_client: MMPClient
                 ) -> None:
        """Create a MuscleManagerHandler.

        Args:
            instance_id: The name of the instance we're logging for.
            level: The initial log level to use.
            mmp_client: The MMP Client to submit log messages to.
        """
        super().__init__(level)
        self._instance_id = instance_id
        self._manager = mmp_client

    def emit(self, record: logging.LogRecord) -> None:
        message = LogMessage(self._instance_id, Timestamp(record.created),
                             LogLevel.from_python_level(record.levelno),
                             self.format(record))
        self._manager.submit_log_message(message)



from unittest.mock import patch

from libmuscle.manager.profile_database import ProfileDatabase
from libmuscle.manager.profile_store import ProfileStore
from libmuscle.planner.planner import ResourceAssignment
from libmuscle.profiling import (
        ProfileEvent, ProfileEventType, ProfileTimestamp)

from libmuscle.test.conftest import on_node_resources as onr

from ymmsl import Operator, Port, Reference

import pytest

from pathlib import Path


@pytest.fixture
def db_file(tmp_path) -> Path:
    with patch('libmuscle.manager.profile_store._SYNCHED', True):
        store = ProfileStore(tmp_path)
        t1 = ProfileTimestamp()

        store.store_instances([Reference('instance1'), Reference('instance2')])

        resources1 = ResourceAssignment([
            onr('node001', {0, 1}), onr('node002', {0, 1})])

        resources2 = ResourceAssignment([
            onr('node001', {0}), onr('node002', {0, 1, 2})])

        store.store_resources({
            Reference('instance1'): resources1,
            Reference('instance2'): resources2})

        def t(offset: int) -> ProfileTimestamp:
            return ProfileTimestamp(t1.nanoseconds + offset)

        e1 = [
                ProfileEvent(ProfileEventType.REGISTER, t(0), t(1000)),
                ProfileEvent(ProfileEventType.CONNECT, t(1000), t(2000)),
                ProfileEvent(
                    ProfileEventType.SEND, t(2000), t(2100),
                    Port('out', Operator.O_I), None, None, 1000000, 0.0),
                ProfileEvent(
                    ProfileEventType.SEND, t(2200), t(2400),
                    Port('out', Operator.O_I), None, None, 1000000, 0.0),
                ProfileEvent(
                    ProfileEventType.SEND, t(2600), t(2900),
                    Port('out', Operator.O_I), None, None, 1000000, 0.0),
                ProfileEvent(ProfileEventType.SHUTDOWN_WAIT, t(10000), t(11000)),
                ProfileEvent(ProfileEventType.DEREGISTER, t(11000), t(11100))]

        store.add_events(Reference('instance1'), e1)

        e2 = [
                ProfileEvent(ProfileEventType.REGISTER, t(0), t(800)),
                ProfileEvent(ProfileEventType.CONNECT, t(800), t(1600)),
                ProfileEvent(
                    ProfileEventType.RECEIVE, t(2000), t(2100),
                    Port('in', Operator.O_I), None, None, 1000000, 0.0),
                ProfileEvent(
                    ProfileEventType.RECEIVE_WAIT, t(2000), t(2090),
                    Port('in', Operator.O_I), None, None, 1000000, 0.0),
                ProfileEvent(
                    ProfileEventType.RECEIVE, t(2200), t(2400),
                    Port('in', Operator.O_I), None, None, 1000000, 0.0),
                ProfileEvent(
                    ProfileEventType.RECEIVE_WAIT, t(2200), t(2380),
                    Port('in', Operator.O_I), None, None, 1000000, 0.0),
                ProfileEvent(
                    ProfileEventType.RECEIVE, t(2600), t(2900),
                    Port('in', Operator.O_I), None, None, 1000000, 0.0),
                ProfileEvent(
                    ProfileEventType.RECEIVE_WAIT, t(2600), t(2870),
                    Port('in', Operator.O_I), None, None, 1000000, 0.0),
                ProfileEvent(ProfileEventType.SHUTDOWN_WAIT, t(10000), t(11000)),
                ProfileEvent(ProfileEventType.DEREGISTER, t(11000), t(11100))]

        store.add_events(Reference('instance2'), e2)

        db_file = store._db_file
        store.shutdown()

    return db_file


def test_instance_stats(db_file):
    with ProfileDatabase(db_file) as db:
        instances, run_time, comm_time, wait_time = db.instance_stats()
        assert set(instances) == {'instance1', 'instance2'}
        i1 = instances.index('instance1')
        assert run_time[i1] == pytest.approx(7400.0 * 1e-9)
        assert comm_time[i1] == pytest.approx(600.0 * 1e-9)
        assert wait_time[i1] == 0.0
        i2 = instances.index('instance2')
        assert run_time[i2] == pytest.approx(7800.0 * 1e-9)
        assert comm_time[i2] == pytest.approx(60.0 * 1e-9)
        assert wait_time[i2] == pytest.approx(540.0 * 1e-9)


def test_resource_stats(db_file):
    with ProfileDatabase(db_file) as db:
        stats = db.resource_stats()
        assert set(stats.keys()) == set([
            'node001:0', 'node001:1',
            'node002:0', 'node002:1', 'node002:2'])

        assert stats['node001:0'] == {
                'instance1': 8000.0 * 1e-9,
                'instance2': 7860.0 * 1e-9}

        assert stats['node001:1'] == {
                'instance1': 8000.0 * 1e-9}

        assert stats['node002:0'] == {
                'instance1': 8000.0 * 1e-9,
                'instance2': 7860.0 * 1e-9}

        assert stats['node002:1'] == {
                'instance1': 8000.0 * 1e-9,
                'instance2': 7860.0 * 1e-9}

        assert stats['node002:2'] == {
                'instance2': 7860.0 * 1e-9}


def test_time_taken(db_file):
    with ProfileDatabase(db_file) as db:
        assert 1000.0 == db.time_taken(etype='REGISTER', instance='instance1')
        assert 100.0 == db.time_taken(etype='DEREGISTER')
        assert 11100.0 == db.time_taken(
                etype='REGISTER', instance='instance1', etype2='DEREGISTER')
        assert 10000.0 == db.time_taken(
                etype='REGISTER', instance='instance1', time='stop',
                etype2='DEREGISTER', time2='start')
        assert 200.0 == db.time_taken(etype='SEND')
        assert 200.0 == db.time_taken(etype='DEREGISTER', aggregate='sum')
        assert 600.0 == db.time_taken(etype='SEND', aggregate='sum')



from typing import Dict, List, Optional, Tuple

from ymmsl import Identifier, Operator

from libmuscle.peer_info import PeerInfo
from libmuscle.port import Port


class PortManager:
    """Manages sending and receiving ports of the current instance."""
    def __init__(
            self, index: List[int], declared_ports: Optional[Dict[Operator, List[str]]]
            ) -> None:
        """Create a PortManager.

        Args:
            index: The index of this instance.
            declared_ports: The ports this instance has, as declared by the
                    user when creating the Instance object.
        """
        self._index = index
        self._ports: Dict[str, Port] = {}
        self._declared_ports = declared_ports

    def connect_ports(self, peer_info: PeerInfo) -> None:
        """Connect the ports to their peers.

        This is the second stage in the simulation wiring process.

        Peers here are instances, and the information about them received
        from the manager is in peer_info. We are going to create a set of
        Port objects here, one for each of our ports.

        If the user gave us a set of ports (i.e. they were declared in the
        code) then we'll create a Port object for each of those, with
        information about the attached conduit / peer (if any) from
        peer_info. If the user did not give us any ports (which is legal),
        then the ports will be created entirely from the information
        received from the manager.  The user then has to use
        Instance.list_ports() to see what ports they got, and do something
        with them.

        Args:
            peer_info: Information about our peers from the manager.
        """
        self._muscle_settings_in = self._settings_in_port(peer_info)
        if self._declared_ports:
            self._ports = self._ports_from_declared(peer_info)
        else:
            self._ports = self._ports_from_conduits(peer_info)

    def settings_in_connected(self) -> bool:
        """Returns whether muscle_settings_in is connected."""
        return self._muscle_settings_in.is_connected()

    def list_ports(self) -> Dict[Operator, List[str]]:
        """Returns a description of the ports this PortManager has.

        Returns:
            A dictionary, indexed by Operator, containing lists of
            port names. Operators with no associated ports are not
            included.
        """
        result: Dict[Operator, List[str]] = {}
        for port_name, port in self._ports.items():
            if port.operator not in result:
                result[port.operator] = list()
            result[port.operator].append(port_name)
        return result

    def port_exists(self, port_name: str) -> bool:
        """Returns whether a port with the given name exists.

        Args:
            port_name: Port name to check.

        Returns:
            True iff the port exists
        """
        return port_name in self._ports

    def get_port(self, port_name: str) -> Port:
        """Returns a Port object describing a port with the given name.

        Args:
            port_name: Name of the port to retrieve.

        Returns:
            A Port object for the port
        """
        return self._ports[port_name]

    def restore_message_counts(
            self, port_message_counts: Dict[str, List[int]]) -> None:
        """Restore message counts on all ports.

        Args:
            port_message_counts: The message counts, as a dictionary indexed by
                    port name containing a list of counts, one for each slot.
        """
        for port_name, num_messages in port_message_counts.items():
            if port_name == "muscle_settings_in":
                self._muscle_settings_in.restore_message_counts(num_messages)
            elif port_name in self._ports:
                self._ports[port_name].restore_message_counts(num_messages)
            else:
                raise RuntimeError(f'Unknown port {port_name} in snapshot.'
                                   ' Have your port definitions changed since'
                                   ' the snapshot was taken?')

    def get_message_counts(self) -> Dict[str, List[int]]:
        """Get message counts for all ports on the communicator.

        Returns:
            A dictionary indexed by port name containing a list of counts, one
            for each slot of the corresponding port.
        """
        port_message_counts = {port_name: port.get_message_counts()
                               for port_name, port in self._ports.items()}
        port_message_counts["muscle_settings_in"] = \
            self._muscle_settings_in.get_message_counts()
        return port_message_counts

    def _settings_in_port(self, peer_info: PeerInfo) -> Port:
        """Creates a Port representing muscle_settings_in.

        Args:
            peer_info: Information about our peers from the manager.
        """
        msi = Identifier('muscle_settings_in')
        if peer_info.is_connected(msi):
            sender_component = peer_info.get_peer_ports(msi)[0][:-1]
            return Port(
                    str(msi), Operator.F_INIT, False, True, len(self._index),
                    peer_info.get_peer_dims(sender_component))

        return Port(
                str(msi), Operator.F_INIT, False, False, len(self._index), [])

    def _ports_from_declared(self, peer_info: PeerInfo) -> Dict[str, Port]:
        """Derives port definitions from supplied declaration.

        Args:
            peer_info: Information about our peers from the manager.

        Returns:
            A dictionary keyed by port name containing corresponding Port
                    objects.
        """
        assert self._declared_ports is not None

        ports = dict()
        for operator, port_list in self._declared_ports.items():
            for port_desc in port_list:
                port_name, is_vector = self._split_port_desc(port_desc)
                if port_name.startswith('muscle_'):
                    raise RuntimeError(
                            'Port names starting with "muscle_" are reserved for'
                            ' MUSCLE, please rename port "{}"'.format(port_name))
                port_id = Identifier(port_name)
                is_connected = peer_info.is_connected(port_id)
                if is_connected:
                    peer_ports = peer_info.get_peer_ports(port_id)
                    peer_port = peer_ports[0]
                    peer_component = peer_port[:-1]
                    port_peer_dims = peer_info.get_peer_dims(peer_component)
                    for peer_port in peer_ports[1:]:
                        peer_component = peer_port[:-1]
                        if port_peer_dims != peer_info.get_peer_dims(peer_component):
                            port_strs = ', '.join(map(str, peer_ports))
                            raise RuntimeError(
                                    'Multicast port "{}" is connected to peers with'
                                    ' different dimensions. All peer components that'
                                    ' this port is connected to must have the same'
                                    ' multiplicity. Connected to ports: {}.'.format(
                                        port_name, port_strs))
                else:
                    port_peer_dims = []
                ports[port_name] = Port(
                        port_name, operator, is_vector, is_connected,
                        len(self._index), port_peer_dims)
        return ports

    def _ports_from_conduits(self, peer_info: PeerInfo) -> Dict[str, Port]:
        """Derives port definitions from conduits.

        Args:
            peer_info: Information about our peers from the manager.
        """
        ports = dict()

        def make_port(
                port_id: Identifier, operator: Operator, peer_dims: List[int]) -> None:
            if not str(port_id).startswith('muscle_'):
                ndims = max(0, len(peer_dims) - len(self._index))
                is_vector = (ndims == 1)
                is_connected = peer_info.is_connected(port_id)
                ports[str(port_id)] = Port(
                        str(port_id), operator, is_vector, is_connected,
                        len(self._index), peer_dims)

        for port_id, sender_ref in peer_info.list_incoming_ports():
            peer_dims = peer_info.get_peer_dims(sender_ref[:-1])
            make_port(port_id, Operator.F_INIT, peer_dims)

        for port_id, receiver_refs in peer_info.list_outgoing_ports():
            peer_dims = peer_info.get_peer_dims(receiver_refs[0][:-1])
            make_port(port_id, Operator.O_F, peer_dims)

        return ports

    def _split_port_desc(self, port_desc: str) -> Tuple[str, bool]:
        """Split a port description into its name and dimensionality.

        Expects a port description of the form port_name or
        port_name[], and returns the port name and whether it is a
        vector port.

        Args:
            port_desc: A port description string, as above.
        """
        is_vector = False
        if port_desc.endswith('[]'):
            is_vector = True
            port_desc = port_desc[:-2]

        if port_desc.endswith('[]'):
            raise ValueError(('Port description "{}" is invalid: ports can'
                              ' have at most one dimension.').format(
                                  port_desc))

        return port_desc, is_vector



from typing import Optional, Tuple

from libmuscle.profiling import ProfileTimestamp


ProfileData = Tuple[ProfileTimestamp, ProfileTimestamp, ProfileTimestamp]


class TimeoutHandler:
    """Object handling timeouts during :meth:`TransportClient.call`."""

    @property
    def timeout(self) -> float:
        """Timeout (in seconds) after which :meth:`on_timeout` is called."""
        raise NotImplementedError()     # pragma: no cover

    def on_timeout(self) -> None:
        """Callback when :attr:`timeout` seconds have passed without a response from the
        peer.
        """
        raise NotImplementedError()     # pragma: no cover

    def on_receive(self) -> None:
        """Callback when receiving a response from the peer.

        Note: this method is only called when the request has timed out.
        """
        raise NotImplementedError()     # pragma: no cover


class TransportClient:
    """A client that connects to an MCP server.

    This is a base class for MCP Transport Clients. An MCP Transport
    Client connects to an MCP Transport Server over some communication
    protocol, requests messages from it, and returns responses.
    """
    @staticmethod
    def can_connect_to(location: str) -> bool:
        """Whether this client class can connect to the given location.

        Args:
            location: The location to potentially connect to.

        Returns:
            True iff this class can connect to this location.
        """
        raise NotImplementedError()     # pragma: no cover

    def call(self, request: bytes, timeout_handler: Optional[TimeoutHandler] = None
             ) -> Tuple[bytes, ProfileData]:
        """Send a request to the server and receive the response.

        This is a blocking call. Besides the result, this function
        returns a tuple with three timestamps (floats in seconds since
        the epoch). These were taken when the function was first called,
        when data became available and the transfer started, and when
        the transfer stopped.

        Args:
            request: The request to send
            timeout_handler: Optional timeout handler. This is used for communication
                deadlock detection.

        Returns:
            The received response, and the timestamps
        """
        raise NotImplementedError()     # pragma: no cover

    def close(self) -> None:
        """Closes this client.

        This closes any connections this client has and/or performs
        other shutdown activities as needed.
        """
        raise NotImplementedError()     # pragma: no cover



from libmuscle.profiling import (
        ProfileEvent, ProfileEventType, ProfileTimestamp)
from libmuscle.manager.profile_store import ProfileStore
from ymmsl import Operator, Port, Reference

import sqlite3
from unittest.mock import patch


def test_create_profile_store(tmp_path):
    db = ProfileStore(tmp_path)
    db.shutdown()

    db_path = tmp_path / 'performance.sqlite'
    conn = sqlite3.connect(db_path, isolation_level=None)
    cur = conn.cursor()
    cur.execute("BEGIN TRANSACTION")
    cur.execute("SELECT major_version, minor_version FROM muscle3_format")
    major, minor = cur.fetchone()
    assert major == 1
    assert minor == 1

    cur.execute("SELECT oid, name FROM event_types")
    etypes = cur.fetchall()
    assert len(etypes) == len(ProfileEventType)

    cur.execute("SELECT oid, name FROM port_operators")
    opers = cur.fetchall()
    assert len(opers) == len(Operator)

    cur.execute("SELECT oid, name FROM instances")
    instances = cur.fetchall()
    assert len(instances) == 0

    cur.execute(
            "SELECT instance_oid, event_type_oid, start_time, stop_time, port_name,"
            "       port_operator_oid, port_length, slot, message_number, message_size,"
            "       message_timestamp FROM events")
    events = cur.fetchall()
    assert len(events) == 0

    cur.execute("COMMIT")
    cur.close()
    conn.close()


@patch('libmuscle.manager.profile_store._SYNCHED', True)
def test_add_events(tmp_path):
    db = ProfileStore(tmp_path)

    db_path = tmp_path / 'performance.sqlite'
    conn = sqlite3.connect(db_path, isolation_level=None)
    cur = conn.cursor()

    events = [
            ProfileEvent(
                ProfileEventType.REGISTER, ProfileTimestamp(0),
                ProfileTimestamp(1000)),
            ProfileEvent(
                ProfileEventType.SEND, ProfileTimestamp(800),
                ProfileTimestamp(812),
                Port('out_port', Operator.O_I), 10, 3, 67, 12345, 13.42),
            ProfileEvent(
                ProfileEventType.DEREGISTER, ProfileTimestamp(1000000000000),
                ProfileTimestamp(1100000000000))]

    def check_send_event(instance):
        cur.execute("BEGIN TRANSACTION")
        cur.execute(
                "SELECT *"
                " FROM events AS e, instances AS i, event_types AS et,"
                "      port_operators AS o"
                " WHERE e.instance_oid = i.oid AND e.event_type_oid = et.oid"
                " AND e.port_operator_oid = o.oid AND i.name = 'instance[0]'"
                " AND et.name = (?)", (ProfileEventType.SEND.name,))
        events2 = cur.fetchall()

        assert len(events2) == 1
        e = events2[0]
        assert e[1:11] == (
                ProfileEventType.SEND.value, 800, 812, 'out_port',
                Operator.O_I.value, 10, 3, 67, 12345, 13.42)
        assert e[12] == 'instance[0]'
        assert e[14] == 'SEND'
        assert e[16] == 'O_I'

        cur.execute("COMMIT")

    db.store_instances([Reference('instance[0]'), Reference('instance[1]')])

    db.add_events(Reference('instance[0]'), events)
    check_send_event('instance[0]')

    db.add_events(Reference('instance[1]'), events)
    check_send_event('instance[1]')

    def check_register_event(typ, start, stop):
        cur.execute("BEGIN TRANSACTION")
        cur.execute(
                "SELECT i.name, e.start_time, e.stop_time"
                " FROM events AS e, instances AS i, event_types AS et"
                " WHERE e.instance_oid = i.oid AND e.event_type_oid = et.oid"
                " AND et.name = ?", (typ,))

        events2 = cur.fetchall()
        cur.execute("COMMIT")

        assert len(events2) == 2
        assert set(events2) == {
                ('instance[0]', start, stop), ('instance[1]', start, stop)}

    check_register_event('REGISTER', 0, 1000)
    check_register_event('DEREGISTER', 1000000000000, 1100000000000)

    cur.close()
    conn.close()
    db.shutdown()



from dataclasses import dataclass, field
from datetime import datetime
from enum import Flag, auto
from functools import lru_cache
from itertools import chain, zip_longest
from operator import attrgetter
from pathlib import Path
from queue import Queue
from threading import Thread
from typing import Dict, Optional, Set, FrozenSet, List, Tuple, TypeVar

from ymmsl import (
        Reference, Model, Identifier, Implementation, save,
        PartialConfiguration)

from libmuscle.manager.topology_store import TopologyStore
from libmuscle.snapshot import SnapshotMetadata


_MAX_FILE_EXISTS_CHECK = 100

_SnapshotDictType = Dict[Reference, List["SnapshotNode"]]
_ConnectionType = Tuple[Identifier, Identifier, "_ConnectionInfo"]
_QueueItemType = Optional[Tuple[Reference, SnapshotMetadata]]
_T = TypeVar("_T")

# this snapshot is used as a placeholder for restarting from scratch
_NULL_SNAPSHOT = SnapshotMetadata(["Instance start"], 0, 0, None, {}, True, '')


def safe_get(lst: List[_T], index: int, default: _T) -> _T:
    """Get an item from the list, returning default when it does not exist.

    Args:
        lst: List to get the item from
        index: Which item to get, should be >= 0
        default: Value to return when hitting an IndexError
    """
    try:
        return lst[index]
    except IndexError:
        return default


class _ConnectionInfo(Flag):
    SELF_IS_SENDING = auto()
    SELF_IS_VECTOR = auto()
    PEER_IS_VECTOR = auto()


def calc_consistency(
        num1: int, num2: int, first_is_sent: bool, num2_is_restart: bool
        ) -> bool:
    """Calculate consistency of message counts.

    Args:
        num1: message count of instance 1
        num2: message count of instance 2
        first_is_sent: True iff instance 1 is sending messages over this conduit
        num2_is_restart: True iff the snapshot of num2 is a full restart

    Returns:
        True iff the two message counts are consistent
    """
    return (num1 == num2 or                             # strong
            num1 + 1 == num2 and first_is_sent or       # weak (1 = sent)
            # weak (2 = sent) - only allow if num2 is not a restart
            num2 + 1 == num1 and not first_is_sent and not num2_is_restart)


def calc_consistency_list(
        num1: List[int], num2: List[int], first_is_sent: bool,
        num2_is_restart: bool) -> bool:
    """Calculate consistency of message counts.

    Args:
        num1: message count of instance 1
        num2: message count of instance 2
        first_is_sent: True iff instance 1 is sending messages over this conduit
        num2_is_restart: True iff the snapshot of num2 is a full restart

    Returns:
        True iff the two message counts are consistent
    """
    if first_is_sent:
        allow_weak = True
        slot_iter = zip_longest(num1, num2, fillvalue=0)
    else:
        allow_weak = not num2_is_restart
        slot_iter = zip_longest(num2, num1, fillvalue=0)
    return all(slot_sent == slot_received or                    # strong
               slot_sent + 1 == slot_received and allow_weak    # weak
               for slot_sent, slot_received in slot_iter)


@dataclass
class SnapshotNode:
    """Represents a node in the snapshot graph.

    Attributes:
        num: The number of the snapshot. Unique for this instance. Later
            snapshots always have a higher num.
        instance: Which instance this is a snapshot of.
        snapshot: The snapshot metadata reported by the instance.
        peers: The set of peers that the instance is connected to.
        consistent_peers: Keeps track of snapshots per peer that are consistent
            with this one.
    """
    num: int
    instance: Reference
    snapshot: SnapshotMetadata
    peers: FrozenSet[Reference]
    consistent_peers: Dict[Reference, List["SnapshotNode"]] = field(
            default_factory=dict, repr=False)

    def __hash__(self) -> int:
        return object.__hash__(self)

    @property
    def consistent(self) -> bool:
        """Returns True iff there is a consistent checkpoint with all peers.
        """
        return self.consistent_peers.keys() == self.peers

    def do_consistency_check(
            self,
            peer_node: "SnapshotNode",
            connections: List[_ConnectionType]) -> bool:
        """Check if the snapshot of the peer is consistent with us.

        When the peer snapshot is consistent, adds it to our list of consistent
        peer snapshots (in :attr:`consistent_peers`) and vice versa.

        Args:
            peer_node: Snapshot of one of our peers
            connections: All connections from our instance to the peer instance

        Returns:
            True iff the peer snapshot is consistent with ours.
        """
        i_snapshot = self.snapshot
        p_snapshot = peer_node.snapshot
        peer_is_restart = p_snapshot is _NULL_SNAPSHOT
        for connection in connections:
            i_port, p_port, conn = connection
            is_sending = bool(conn & _ConnectionInfo.SELF_IS_SENDING)
            i_msg_counts = i_snapshot.port_message_counts.get(str(i_port), [])
            p_msg_counts = p_snapshot.port_message_counts.get(str(p_port), [])
            if conn & _ConnectionInfo.SELF_IS_VECTOR:
                slot = int(peer_node.instance[-1])
                consistent = calc_consistency(
                        safe_get(i_msg_counts, slot, 0),
                        safe_get(p_msg_counts, 0, 0),
                        is_sending, peer_is_restart)
            elif conn & _ConnectionInfo.PEER_IS_VECTOR:
                slot = int(self.instance[-1])
                consistent = calc_consistency(
                        safe_get(i_msg_counts, 0, 0),
                        safe_get(p_msg_counts, slot, 0),
                        is_sending, peer_is_restart)
            else:
                consistent = calc_consistency_list(
                        i_msg_counts, p_msg_counts, is_sending, peer_is_restart)
            if not consistent:
                return False
        self.consistent_peers.setdefault(
                peer_node.instance, []).append(peer_node)
        peer_node.consistent_peers.setdefault(
                self.instance, []).append(self)
        return True


class SnapshotRegistry(Thread):
    """Registry of all snapshots taken by instances.

    Current snapshots are stored in a graph. Every node represents a snapshot
    taken by an instance (see :class:`SnapshotNode`). When snapshots from peer
    instances are consistent, the nodes are connected to each other.

    This class manages the snapshot nodes. New snapshots are registered through
    :meth:`register_snapshot`.
    """

    def __init__(
            self, config: PartialConfiguration, snapshot_folder: Path,
            topology_store: TopologyStore) -> None:
        """Create a snapshot graph using provided configuration.

        Args:
            config: ymmsl configuration describing the workflow.
        """
        super().__init__(name='SnapshotRegistry')

        if config.model is None or not isinstance(config.model, Model):
            raise ValueError('The yMMSL experiment description does not'
                             ' contain a (complete) model section, so there'
                             ' is nothing to run!')
        self._configuration = config
        self._model = config.model
        self._snapshot_folder = snapshot_folder
        self._topology_store = topology_store

        self._queue: Queue[_QueueItemType] = Queue()
        self._snapshots: _SnapshotDictType = {}

        self._instances: Set[Reference] = set()
        for component in config.model.components:
            self._instances.update(component.instances())

        # Create snapshot nodes for starting from scratch
        for instance in self._instances:
            self.register_snapshot(instance, _NULL_SNAPSHOT)

    def register_snapshot(
            self, instance: Reference, snapshot: SnapshotMetadata) -> None:
        """Register a new snapshot.

        Args:
            instance: The instance that created the snapshot
            snapshot: Metadata describing the snapshot
        """
        self._queue.put((instance, snapshot))

    def run(self) -> None:
        """Code executed in a separate thread
        """
        while True:
            item = self._queue.get()
            if item is None:
                return
            self._add_snapshot(*item)

    def shutdown(self) -> None:
        """Stop the snapshot registry thread
        """
        self._queue.put(None)

    def _add_snapshot(
            self, instance: Reference, snapshot: SnapshotMetadata) -> None:
        """Register a new snapshot.

        Args:
            instance: The instance that created the snapshot
            snapshot: Metadata describing the snapshot
        """
        stateful_peers = self._get_peers(instance)

        i_snapshots = self._snapshots.setdefault(instance, [])
        # get next number of the snapshot
        num = 1 if not i_snapshots else i_snapshots[-1].num + 1
        snapshotnode = SnapshotNode(num, instance, snapshot, stateful_peers)
        i_snapshots.append(snapshotnode)

        # check consistency with all peers
        for peer in stateful_peers:
            for peer_snapshot in self._snapshots.get(peer, []):
                snapshotnode.do_consistency_check(
                        peer_snapshot, self._get_connections(instance, peer))

        # finally, check if this snapshotnode is now part of a workflow snapshot
        if snapshot is not _NULL_SNAPSHOT:
            self._save_workflow_snapshot(snapshotnode)

    def _save_workflow_snapshot(self, snapshotnode: SnapshotNode) -> None:
        """Save snapshot if a workflow snapshot exists with the provided node.

        Args:
            snapshotnode: The snapshot node that must be part of the workflow
                snapshot.
        """
        workflow_snapshots = self._get_workflow_snapshots(snapshotnode)
        for workflow_snapshot in workflow_snapshots:
            self._write_snapshot_ymmsl(workflow_snapshot)
        self._cleanup_snapshots(workflow_snapshots)

    def _get_workflow_snapshots(
            self, snapshot: SnapshotNode) -> List[List[SnapshotNode]]:
        """Return all workflow snapshots which contain the provided node.

        Args:
            snapshotnode: The snapshot node that must be part of the workflow
                snapshot.

        Returns:
            List of workflow snapshots. Each workflow snapshot is a list of
            instance snapshot nodes.
        """
        if not snapshot.consistent:
            return []

        # Instances that don't have a snapshot node chosen yet:
        instances_to_cover = list(self._instances - {snapshot.instance})
        # Allowed snapshots per instance. This is updated during the heuristic
        # to further restrict the sets of snapshots as peer snapshots are
        # selected.
        # First restriction is that the snapshots have to be locally consistent.
        allowed_snapshots: Dict[Reference, FrozenSet[SnapshotNode]] = {}
        for instance in instances_to_cover:
            allowed_snapshots[instance] = frozenset(
                    i_snapshot
                    for i_snapshot in self._snapshots.get(instance, [])
                    if i_snapshot.consistent)
            if not allowed_snapshots[instance]:
                # there cannot be a workflow snapshot if this instance has no
                # consistent snapshot nodes
                return []
        instance = snapshot.instance
        allowed_snapshots[instance] = frozenset({snapshot})

        def num_allowed_snapshots(instance: Reference) -> int:
            """Get number of allowed snapshots at this point for this instance.

            The allowed snapshots are those that are consistent with all
            selected snapshots at this point in the heuristic.
            """
            return len(allowed_snapshots[instance])

        # Do a full, depth-first search for all workflow snapshots
        # ========================================================

        workflow_snapshots = []
        selected_snapshots = [snapshot]
        # This stack stores history of allowed_snapshots and enables roll back
        stack: List[Dict[Reference, FrozenSet[SnapshotNode]]] = []

        # Update allowed_snapshots for peers of the selected snapshot
        for peer, snapshots in snapshot.consistent_peers.items():
            intersection = allowed_snapshots[peer].intersection(snapshots)
            if not intersection:
                return []
            allowed_snapshots[peer] = intersection

        while True:
            # 1. Select most constrained instance
            #
            # Note: we're only interested in the instance with the least allowed
            # snapshots. Better performance may be possible by not doing a full
            # sort, but it should be tested. Expectation is that
            # instances_to_cover remains mostly sorted (as the only counts that
            # are changing are for peers of the previous selected instance).
            # Python's sort algorithm is O(N) when the list is already sorted
            # (which is the same as max()).
            #
            # We cannot use a priority queue (heapq) because
            # num_allowed_snapshots is changing every iteration.
            instances_to_cover.sort(key=num_allowed_snapshots, reverse=True)
            instance = instances_to_cover.pop()

            # 2. Select the oldest snapshot of this instance
            snapshot = min(allowed_snapshots[instance], key=attrgetter('num'))
            selected_snapshots.append(snapshot)
            # A shallow copy is ok: the values are immutable frozensets
            stack.append(allowed_snapshots.copy())

            # 3. Update allowed snapshots based on the newly selected
            allowed_snapshots[instance] = frozenset({snapshot})
            for peer, snapshots in snapshot.consistent_peers.items():
                intersection = allowed_snapshots[peer].intersection(snapshots)
                if not intersection:
                    break  # roll back
                allowed_snapshots[peer] = intersection
            else:
                # 4. Selected snapshot is okay to explore further
                if instances_to_cover:
                    # 4a. There are still instance to cover, return to the start
                    #     of the while loop.
                    continue
                # 4b. We have found a complete workflow snapshot
                workflow_snapshots.append(selected_snapshots.copy())
                # Next: perform a roll-back to continue the search

            # 5. Roll back
            # stop when selected_snapshots only contains the one we forced to be
            # part of the workflow snapshot
            while len(selected_snapshots) > 1:
                snapshot = selected_snapshots.pop()
                instance = snapshot.instance
                instances_to_cover.append(instance)
                allowed_snapshots = stack.pop()
                intersection = allowed_snapshots[instance] - {snapshot}
                allowed_snapshots[instance] = intersection
                if intersection:
                    # We have a valid next snapshot to try for this instance
                    break
                # No allowed_snapshots, try another roll back
            else:
                # Exhausted all roll back possibilities, so we are done now
                return workflow_snapshots

    def _write_snapshot_ymmsl(
            self, selected_snapshots: List[SnapshotNode]) -> None:
        """Write the snapshot ymmsl file to the snapshot folder.

        Args:
            selected_snapshots: All snapshot nodes of the workflow snapshot.
        """
        now = datetime.now()
        config = self._generate_snapshot_config(selected_snapshots, now)
        time = now.strftime('%Y%m%d_%H%M%S')
        for i in range(_MAX_FILE_EXISTS_CHECK):
            if i:
                snapshot_filename = f'snapshot_{time}_{i}.ymmsl'
            else:
                snapshot_filename = f'snapshot_{time}.ymmsl'
            savepath = self._snapshot_folder / snapshot_filename
            if not savepath.exists():
                save(config, savepath)
                return
        raise RuntimeError('Could not find an available filename for storing'
                           f' the next workflow snapshot: {savepath} already'
                           ' exists.')

    def _generate_snapshot_config(
                self, selected_snapshots: List[SnapshotNode], now: datetime
                ) -> PartialConfiguration:
        """Generate ymmsl configuration for snapshot file
        """
        selected_snapshots.sort(key=attrgetter('instance'))
        resume = {}
        for node in selected_snapshots:
            if node.snapshot is not _NULL_SNAPSHOT:
                # Only store resume information when it is an actual snapshot
                # created by the instance. Otherwise the instance can just be
                # restarted from the beginning.
                resume[node.instance] = Path(node.snapshot.snapshot_filename)
        description = self._generate_description(selected_snapshots, now)
        return PartialConfiguration(resume=resume, description=description)

    def _generate_description(
            self, selected_snapshots: List[SnapshotNode], now: datetime) -> str:
        """Generate a human-readable description of the workflow snapshot.
        """
        triggers: Dict[str, List[str]] = {}
        component_info = []
        max_instance_len = len('Instance ')
        for node in selected_snapshots:
            for trigger in node.snapshot.triggers:
                triggers.setdefault(trigger, []).append(str(node.instance))
            component_info.append((
                    str(node.instance),
                    f'{node.snapshot.timestamp:<11.6g}',
                    f'{node.snapshot.wallclock_time:<11.6g}',
                    ("Intermediate", "Final")[node.snapshot.is_final_snapshot]))
            max_instance_len = max(max_instance_len, len(str(node.instance)))
        instance_with_padding = 'Instance'.ljust(max_instance_len)
        component_table = [
                f'{instance_with_padding} t           Wallclock time  Type',
                f'{"-" * (max_instance_len + 41)}']
        component_table += [
                f'{name.ljust(max_instance_len)} {timestamp} {walltime}'
                f'     {typ}'
                for name, timestamp, walltime, typ in component_info]
        return (f'Workflow snapshot for {self._model.name}'
                f' taken on {now.strftime("%Y-%m-%d %H:%M:%S")}.\n'
                'Snapshot triggers:\n' +
                '\n'.join(f'- {trigger} ({", ".join(triggers[trigger])})'
                          for trigger in sorted(triggers)) +
                '\n\n' +
                '\n'.join(component_table) + '\n')

    def _cleanup_snapshots(
            self, workflow_snapshots: List[List[SnapshotNode]]) -> None:
        """Remove all snapshots that are older than the selected snapshots.

        Args:
            selected_snapshots: All snapshot nodes of a workflow snapshot
        """
        if not workflow_snapshots:
            return

        # Find the newest snapshots per instance
        newest_snapshots = {snapshot.instance: snapshot
                            for snapshot in workflow_snapshots[0]}
        for workflow_snapshot in workflow_snapshots[1:]:
            for snapshot in workflow_snapshot:
                if newest_snapshots[snapshot.instance].num < snapshot.num:
                    newest_snapshots[snapshot.instance] = snapshot

        # Remove all snapshots that are older than the newest snapshots
        removed_snapshots: Set[SnapshotNode] = set()
        for snapshot in newest_snapshots.values():
            all_snapshots = self._snapshots[snapshot.instance]
            idx = all_snapshots.index(snapshot)
            self._snapshots[snapshot.instance] = all_snapshots[idx:]
            removed_snapshots.update(all_snapshots[:idx])

        # Remove all references in SnapshotNode.peer_snapshot to the snapshots
        # that are cleaned up
        for snapshot in removed_snapshots:
            for peer_snapshot in chain.from_iterable(
                    snapshot.consistent_peers.values()):
                if peer_snapshot in removed_snapshots:
                    # snapshot is removed anyway, no need to update references
                    continue
                # peer_snapshot is still there, remove reference to us
                peer_snapshot.consistent_peers[snapshot.instance].remove(
                        snapshot)

    @lru_cache(maxsize=None)
    def _get_peers(self, instance: Reference) -> FrozenSet[Reference]:
        """Return the set of peers for the given instance.

        Note: instance is assumed to contain the full index, not just the
        component name.

        Args:
            instance: Instance to get peers of.

        Returns:
            Frozen set with all peer instances (including their index).
        """
        return frozenset(self._topology_store.get_peer_instances(instance))

    @lru_cache(maxsize=None)
    def _get_connections(self, instance: Reference, peer: Reference
                         ) -> List[_ConnectionType]:
        """Get the list of connections between instance and peer.

        Args:
            instance: Instance reference (including index)
            peer: Peer reference (including index)

        Returns:
            A list of tuples describing all conduits between instance and peer:
                instance_port (Reference): the port of instance that is
                    connected to
                peer_port (Reference): the port on the peer instance
                info (_ConnectionInfo): flag describing the connection. The
                    instance is sending when
                    ``info & _ConnectionInfo.SELF_IS_SENDING`` and receiving
                    otherwise. When the instance port is a vector port and the
                    peer port is a non-vector port, the flag
                    ``_ConnectionInfo.SELF_IS_VECTOR`` is set. In the reverse
                    situation the flag ``_ConnectionInfo.PEER_IS_VECTOR`` is
                    set. When both ports are vector or non-vector, neither flag
                    is set.
        """
        instance_kernel = instance.without_trailing_ints()
        peer_kernel = peer.without_trailing_ints()

        connected_ports: List[_ConnectionType] = []
        for conduit in self._model.conduits:
            if (conduit.sending_component() == instance_kernel and
                    conduit.receiving_component() == peer_kernel):
                conn_type = _ConnectionInfo.SELF_IS_SENDING
            elif (conduit.receiving_component() == instance_kernel and
                    conduit.sending_component() == peer_kernel):
                conn_type = _ConnectionInfo(0)
            else:
                continue
            instance_ndim = (len(instance) - len(instance_kernel))
            peer_ndim = (len(peer) - len(peer_kernel))
            if instance_ndim < peer_ndim:
                conn_type |= _ConnectionInfo.SELF_IS_VECTOR
            if instance_ndim > peer_ndim:
                conn_type |= _ConnectionInfo.PEER_IS_VECTOR
            # we cannot distinguish scalar-scalar vs. vector-vector
            # but it does not matter for this logic :)
            if conn_type & _ConnectionInfo.SELF_IS_SENDING:
                connected_ports.append((
                        conduit.sending_port(),
                        conduit.receiving_port(),
                        conn_type))
            else:
                connected_ports.append((
                        conduit.receiving_port(),
                        conduit.sending_port(),
                        conn_type))
        return connected_ports

    @lru_cache(maxsize=None)
    def _implementation(self, kernel: Reference) -> Optional[Implementation]:
        """Return the implementation of a kernel.

        Args:
            kernel: The kernel to get the implementation for.

        Returns:
            Implementation for the kernel, or None if not provided in the
            configuration.
        """
        implementation = None
        for component in self._model.components:
            if component.name == kernel:
                implementation = component.implementation
        if implementation in self._configuration.implementations:
            return self._configuration.implementations[implementation]
        return None



import enum
import logging
import multiprocessing as mp
import os
from pathlib import Path
import traceback
from typing import Dict, Optional

from ymmsl import BaseEnv, Implementation, Reference, ResourceRequirements

from libmuscle.planner.planner import ResourceAssignment


class ProcessStatus(enum.Enum):
    """Status of a process (instance)."""
    STARTED = 0
    RUNNING = 1
    SUCCESS = 2
    ERROR = 3
    CANCELED = 4

    def is_finished(self) -> bool:
        """Returns whether the job is finished.

        Canceled jobs are considered finished, even if they've never
        run.
        """
        return self in (
                ProcessStatus.SUCCESS, ProcessStatus.ERROR,
                ProcessStatus.CANCELED)


class Process:
    """Represents a process.

    Attributes:
        instance: Name of instance this is the process of
        resources: The resources allocated to this process
        status: Current status of the process
        exit_code: Exit code, if status is ERROR
        error_msg: Error message, if status is ERROR
    """
    def __init__(self, instance: Reference, resources: ResourceAssignment) -> None:
        """Create a Process object.

        Args:
            resources: Resources allocated to this process.
        """
        self.instance = instance
        self.resources = resources
        self.status = ProcessStatus.STARTED
        self.exit_code: Optional[int] = None
        self.error_msg: Optional[str] = None


class InstantiatorRequest:
    """Base class for requests to an instantiator."""
    pass


class ShutdownRequest(InstantiatorRequest):
    """Requests shutting down the background process.

    The process will stop once all running processes are done.
    """
    pass


class InstantiationRequest(InstantiatorRequest):
    """Requests instantiating a new process.

    Attributes:
        instance: The name of the instance
        implementation: The implementation to start for it
        res_req: The resource requirements for this instance
        resources: The specific resources to start it on
        instance_dir: The main directory for this instance
        work_dir: The directory in which to start it
        stdout_path: Path of file to redirect stdout to
        stderr_path: Path of file to redirect stderr to
    """
    def __init__(
            self, instance: Reference, implementation: Implementation,
            res_req: ResourceRequirements, resources: ResourceAssignment,
            instance_dir: Path, work_dir: Path, stdout_path: Path, stderr_path: Path
            ) -> None:
        """Create an InstantiationRequest.

        Args:
            instance: The name of the instance
            implementation: The implementation to start for it
            res_req: The resource requirements for this instance
            resources: The specific resources to instantiate on
            instance_dir: The main directory for this instance
            work_dir: The directory in which to start it
            stdout_path: Path of file to redirect stdout to
            stderr_path: Path of file to redirect stderr to
        """
        self.instance = instance
        self.implementation = implementation
        self.res_req = res_req
        self.resources = resources
        self.instance_dir = instance_dir
        self.work_dir = work_dir
        self.stdout_path = stdout_path
        self.stderr_path = stderr_path


class CancelAllRequest(InstantiatorRequest):
    """Requests stopping all running processes."""
    pass


class CrashedResult:
    """Signals that the instantiator process crashed."""
    def __init__(self, exception: Optional[BaseException] = None) -> None:
        self.exception = exception


class QueueingLogHandler(logging.Handler):
    """A logging Handler that enqueues records."""
    def __init__(self, queue: mp.Queue) -> None:
        """Create a QueueingLogHandler.

        Args:
            level: The level of this handler
        """
        super().__init__()
        self._queue = queue

    def emit(self, record: logging.LogRecord) -> None:
        """Emit the record by enqueueing it.

        Args:
            record: A log record to enqueue.
        """
        if record.exc_info:
            record.msg += '\n' + ''.join(
                    traceback.format_exception(*record.exc_info))
            record.exc_info = None

        self._queue.put(record)


def reconfigure_logging(queue: mp.Queue) -> None:
    """Reconfigure logging to send to queue.

    This reconfigures the logging subsystem to intercept all log
    messages and send them to the given queue, rather than to the
    previously configured handler.
    """
    root_logger = logging.getLogger()
    for h in list(root_logger.handlers):
        root_logger.removeHandler(h)

    handler = QueueingLogHandler(queue)
    root_logger.addHandler(handler)


def create_instance_env(
        instance: Reference, base_env: BaseEnv, overlay: Dict[str, str]
        ) -> Dict[str, str]:
    """Creates an environment for an instance.

    This takes the current (manager) environment variables and makes
    a copy, then adds or extends it according to the overlay given.

    Keys from overlay that start with + will have the corresponding
    value appended to the matching (by key, without the +) value in
    env, otherwise the value in env gets overwritten.
    """
    if base_env == BaseEnv.LOGIN:
        env = dict()
        for var in ('HOME', 'LOGNAME', 'SHELL', 'TERM', 'USER'):
            if var in os.environ:
                env[var] = os.environ[var]
    else:
        env = os.environ.copy()
        if base_env == BaseEnv.CLEAN:
            # Remove any virtual environment
            #
            # The deactivate command does not get exported by the activate script, so it
            # isn't available to us, as are most of the environment variables it uses to
            # deactivate the venv. So this is a bit of a hack, but it's the best we can
            # do. Note that _OLD_VIRTUAL_PYTHONHOME isn't exported by default either,
            # but we check for it anyway; if we're on a system where PYTHONHOME is
            # needed and the manager needs to be run with an active virtualenv and one
            # of the implementations needs a clean shell, then the user will have to
            # export it manually between activating the virtualenv and starting the
            # manager and things will work.
            vep = env.get('VIRTUAL_ENV_PROMPT')
            if vep is not None:
                if 'PS1' in env:
                    if env['PS1'].startswith(vep):
                        env['PS1'] = env['PS1'][len(vep):]
                del env['VIRTUAL_ENV_PROMPT']

            ovp = env.get('_OLD_VIRTUAL_PYTHONHOME')
            if ovp is not None:
                env['PYTHONHOME'] = ovp
                del env['_OLD_VIRTUAL_PYTHONHOME']

            venv = env.get('VIRTUAL_ENV')
            if venv is not None:
                path = env.get('PATH')
                if path is not None:
                    paths = [p for p in path.split(':') if p != venv + '/bin']
                    env['PATH'] = ':'.join(paths)

                del env['VIRTUAL_ENV']

    env['MUSCLE_INSTANCE'] = str(instance)

    for key, value in overlay.items():
        if key.startswith('+'):
            if key[1:] in env:
                env[key[1:]] += value
            else:
                env[key[1:]] = value
        else:
            env[key] = value
    return env



from copy import copy
import pytest
from typing import Dict, List, Set, Union
from unittest.mock import patch

from ymmsl import Operator, Reference, Settings

from libmuscle.api_guard import APIGuard
from libmuscle.communicator import Message
from libmuscle.mcp.transport_client import ProfileData
from libmuscle.mmp_client import MMPClient
from libmuscle.planner.resources import Core, CoreSet, OnNodeResources, Resources
from libmuscle.port import Port
from libmuscle.profiler import Profiler
from libmuscle.timestamp import Timestamp


@pytest.fixture
def mocked_mmp_client():
    with patch('libmuscle.mmp_client.TcpTransportClient') as mock_ttc:
        yield MMPClient(Reference('component[13]'), ''), mock_ttc.return_value


@pytest.fixture
def message() -> Message:
    return Message(0.0, None, b'test', Settings())


@pytest.fixture
def message2() -> Message:
    return Message(0.0, None, {'test': 17}, Settings())


@pytest.fixture
def guard() -> APIGuard:
    return APIGuard(True)


@pytest.fixture
def profile_data() -> ProfileData:
    return Timestamp(0.0), Timestamp(0.0), Timestamp(0.0)


@pytest.fixture
def mocked_profiler():
    class MockMMPClient:
        def __init__(self):
            self.sent_events = None

        def submit_profile_events(self, events):
            self.sent_events = copy(events)

    mock_mmp_client = MockMMPClient()
    profiler = Profiler(mock_mmp_client)
    yield profiler, mock_mmp_client
    profiler.shutdown()


@pytest.fixture
def profiler_comm_int_10ms():
    with patch('libmuscle.profiler._COMMUNICATION_INTERVAL', 0.01):
        yield None


@pytest.fixture
def declared_ports():
    return {
            Operator.F_INIT: ['in', 'not_connected'],
            Operator.O_I: ['out_v', 'out_r'],
            Operator.S: ['in_v', 'in_r', 'not_connected_v'],
            Operator.O_F: ['out']}


@pytest.fixture
def mock_ports():
    in_port = Port('in', Operator.F_INIT, False, True, 0, [])
    nc_port = Port('not_connected', Operator.F_INIT, False, False, 0, [])
    outv_port = Port('out_v', Operator.O_I, True, True, 0, [13])
    outr_port = Port('out_r', Operator.O_I, True, True, 0, [])
    inv_port = Port('in_v', Operator.S, True, True, 0, [13])
    inr_port = Port('in_r', Operator.S, True, True, 0, [])
    ncv_port = Port('not_connected_v', Operator.S, True, False, 0, [])
    out_port = Port('out', Operator.O_F, False, True, 0, [])

    return {
            'in': in_port, 'not_connected': nc_port, 'out_v': outv_port,
            'out_r': outr_port, 'in_v': inv_port, 'in_r': inr_port,
            'not_connected_v': ncv_port, 'out': out_port}


@pytest.fixture
def connected_port_manager(port_manager, declared_ports, mock_ports):

    def get_port(name):
        return mock_ports[name]

    def port_exists(name):
        return name in mock_ports

    port_manager.get_port = get_port
    port_manager.list_ports.return_value = declared_ports
    port_manager.port_exists = port_exists
    return port_manager


def core(hwthread_id: int) -> Core:
    """Helper that defines a core with the given core and hwthread id."""
    return Core(hwthread_id, {hwthread_id})


def on_node_resources(node_name: str, cores: Union[int, Set[int]]) -> OnNodeResources:
    """Helper that defines resources on a node from the name and a CPU core."""
    if isinstance(cores, int):
        cores = {cores}
    return OnNodeResources(node_name, CoreSet([Core(core, {core}) for core in cores]))


def resources(node_resources: Dict[str, List[Core]]) -> Resources:
    """Helper that defines a Resources from a dict."""
    return Resources([
        OnNodeResources(node_name, CoreSet(cores))
        for node_name, cores in node_resources.items()])



import sqlite3
from pathlib import Path
from typing import cast, List, Optional, Tuple

import numpy as np
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib import pyplot as plt
from matplotlib.patches import Rectangle

from libmuscle import ProfileDatabase


def plot_instances(performance_file: Path) -> None:
    """Plot instances

    Reads data from the database and makes a bar chart with for each
    instance how much time it spend running, waiting, and
    communicating.

    This won't actually show the plot until show_plots() is called.

    Args:
        performance_file: Database to load data from.
    """
    with ProfileDatabase(performance_file) as db:
        instances, compute, transfer, wait = db.instance_stats()

    fig, ax = plt.subplots()

    width = 0.5
    bottom = np.zeros(len(instances))
    ax.bar(instances, compute, width, label='Compute', bottom=bottom)
    bottom += compute
    ax.bar(instances, transfer, width, label='Transfer', bottom=bottom)
    bottom += transfer
    ax.bar(instances, wait, width, label='Wait', bottom=bottom)
    ax.set_title('Simulation component time breakdown')
    ax.set_xlabel('Instance')
    ax.tick_params(axis='x', labelrotation = 45)
    for label in ax.xaxis.get_ticklabels():
        label.set_horizontalalignment('right')
    ax.set_ylabel('Total time (s)')
    fig.legend(loc='outside right center')
    # bottom=0.3 leaves space for x-axis labels
    # right=0.8 leaves 20% for the legend
    fig.subplots_adjust(bottom=0.3, right=0.8)


def plot_resources(performance_file: Path) -> None:
    """Plot resources

    Reads data from the database and makes a bar chart with for each
    resource (CPU core) which instances ran on it for how long.

    This won't actually show the plot until show_plots() is called.

    Args:
        performance_file: Database to load data from.
    """
    with ProfileDatabase(performance_file) as db:
        stats = db.resource_stats()

    palette = dict()
    next_color = 0
    for core, data in stats.items():
        for instance in data.keys():
            if instance not in palette:
                palette[instance] = f'C{next_color}'
                next_color += 1

    fig, ax = plt.subplots()

    seen_instances = set()
    for i, core in enumerate(sorted(stats.keys())):
        bottom = 0.0
        for instance, time in sorted(stats[core].items(), key=lambda x: -x[1]):
            if instance not in seen_instances:
                label: Optional[str] = instance
                seen_instances.add(instance)
            else:
                label = '_'

            ax.bar(
                    i, time, _BAR_WIDTH,
                    label=label, bottom=bottom, color=palette[instance])
            bottom += time

    ax.set_xticks(range(len(stats)))
    ax.set_xticklabels(stats.keys())

    ax.set_title('Per-core time breakdown')
    ax.set_xlabel('Core')
    ax.tick_params(axis='x', labelrotation = 45)
    for tick_label in ax.xaxis.get_ticklabels():
        tick_label.set_horizontalalignment('right')
    ax.set_ylabel('Total time (s)')
    fig.legend(loc='outside right center')
    # bottom=0.3 leaves space for x-axis labels
    # right=0.8 leaves 20% for the legend
    fig.subplots_adjust(bottom=0.3, right=0.8)


_EVENT_TYPES = (
        'REGISTER', 'CONNECT', 'SHUTDOWN_WAIT', 'DISCONNECT_WAIT', 'SHUTDOWN',
        'DEREGISTER', 'SEND', 'RECEIVE_WAIT', 'RECEIVE_TRANSFER', 'RECEIVE_DECODE')


_EVENT_PALETTE = {
        'REGISTER': '#910f33',
        'CONNECT': '#c85172',
        'SHUTDOWN_WAIT': '#ffdddd',
        'DISCONNECT_WAIT': '#eedddd',
        'SHUTDOWN': '#c85172',
        'DEREGISTER': '#910f33',
        'RECEIVE_WAIT': '#cccccc',
        'RECEIVE_TRANSFER': '#ff7d00',
        'RECEIVE_DECODE': '#ccff00',
        'SEND': '#0095bf'}


_MAX_EVENTS = 5000


_CUTOFF_TEXT = (
        'Warning: data was omitted from the plot in the\n crosshatched'
        ' areas to improve performance.\n Please zoom or pan using the'
        ' tools at the bottom\n of the window to see the missing events.'
        )


_BAR_WIDTH = 0.8


class TimelinePlot:
    """Manages an interactive timeline

    This implements on-demand loading of events as the user pans and
    zooms.
    """
    def __init__(self, performance_file: Path) -> None:
        """Create a TimelinePlot

        This plots the dark gray background bars, and then plots the
        rest on top on demand.

        Args:
            performance_file: The database to plot
        """
        fig, ax = plt.subplots()
        # right=0.8 leaves 20% for the legend
        fig.subplots_adjust(right=0.67)
        self._ax = ax

        # Y axis
        self._cur = sqlite3.connect(performance_file).cursor()
        self._cur.execute("SELECT oid, name FROM instances ORDER BY oid")
        instance_ids, instance_names = zip(*self._cur.fetchall())

        ax.set_yticks(instance_ids)
        ax.set_yticklabels(instance_names)

        # Instances
        self._cur.execute("SELECT MIN(start_time) FROM events")
        self._min_db_time = self._cur.fetchall()[0][0]

        self._cur.execute(
                "SELECT instance_oid, (start_time - ?) * 1e-9"
                " FROM events AS e"
                " JOIN event_types AS et ON (e.event_type_oid = et.oid)"
                " WHERE et.name = 'REGISTER'", (self._min_db_time,))
        begin_times = dict(self._cur.fetchall())

        self._cur.execute(
                "SELECT instance_oid, (stop_time - ?) * 1e-9"
                " FROM events AS e"
                " JOIN event_types AS et ON (e.event_type_oid = et.oid)"
                " WHERE et.name = 'DEREGISTER'", (self._min_db_time,))
        end_times = dict(self._cur.fetchall())

        instances = sorted(begin_times.keys())
        self._instances = instances

        if not begin_times:
            raise RuntimeError(
                    'This database appears to be empty. Did the simulation crash'
                    ' before any data were generated?')

        # Rest of plot
        ax.set_title('Execution timeline')
        ax.set_xlabel('Wallclock time (s)')

        # Background
        running_artist = ax.barh(
                instances,
                [end_times[i] - begin_times[i] for i in instances],
                _BAR_WIDTH,
                left=[begin_times[i] for i in instances],
                label='RUNNING', color='#444444'
                )

        # Initial events plot
        xmin = min(begin_times.values())
        self._global_xmax = max(end_times.values())

        first_cutoff = float('inf')
        self._bars = dict()
        for event_type in _EVENT_TYPES:
            instances, start_times, durations, cutoff = self.get_data(
                    event_type, xmin, self._global_xmax)

            if not instances:
                # Work around https://github.com/matplotlib/matplotlib/issues/21506
                instances = ['']
                start_times = [float('NaN')]
                durations = [float('NaN')]

            self._bars[event_type] = ax.barh(
                    instances, durations, _BAR_WIDTH,
                    label=event_type, left=start_times,
                    color=_EVENT_PALETTE[event_type])
            if cutoff:
                first_cutoff = min(first_cutoff, cutoff)

        # Initial cut-off area
        if first_cutoff != float('inf'):
            self._bars['_CUTOFF'] = ax.barh(
                    self._instances, self._global_xmax - first_cutoff, _BAR_WIDTH,
                    label='Not shown', left=first_cutoff,
                    color='#FFFFFF', hatch='x')
            self._cutoff_warning = ax.text(
                    0.02, 0.02, _CUTOFF_TEXT, transform=ax.transAxes, fontsize=12,
                    verticalalignment='bottom', horizontalalignment='left', wrap=True,
                    bbox={
                        'facecolor': '#ffcccc', 'alpha': 0.75})

        ax.set_autoscale_on(True)
        ax.callbacks.connect('xlim_changed', self.update_data)

        ordered_artists = [self._bars[event_type][0] for event_type in _EVENT_TYPES]
        ordered_names = list(_EVENT_TYPES)

        ordered_artists.insert(6, running_artist)
        ordered_names.insert(6, 'RUNNING')

        fig.legend(
                ordered_artists, ordered_names, loc='outside right center',
                bbox_to_anchor=(1.0, 0.6))
        cast(Figure, ax.figure).canvas.draw_idle()

    def close(self) -> None:
        """Closes the database connection"""
        self._cur.close()

    def get_data(
            self, event_type: str, xmin: float, xmax: float
            ) -> Tuple[List[int], List[float], List[float], Optional[float]]:
        """Get events from the database

        Returns three lists with instance oid, start time and duration, and
        the last timepoint returned in case we had too much data to show and
        data got cut off, or None if all matching data was returned.

        Args:
            event_type: Type of events to get
            xmin: Time point after which the event must have stopped
            xmax: Time point before which the event must have started
        """
        self._cur.execute(
                "SELECT"
                "  instance_oid, (start_time - ?) * 1e-9,"
                "  (stop_time - start_time) * 1e-9"
                " FROM events AS e"
                " JOIN event_types AS et ON (e.event_type_oid = et.oid)"
                " WHERE et.name = ?"
                " AND (start_time - ?) * 1e-9 <= ?"
                " AND ? <= (stop_time - ?) * 1e-9"
                " ORDER BY start_time ASC"
                " LIMIT ?",
                (
                    self._min_db_time, event_type, self._min_db_time, xmax,
                    xmin, self._min_db_time, _MAX_EVENTS))
        results = self._cur.fetchall()
        if not results:
            return list(), list(), list(), None

        if len(results) == _MAX_EVENTS:
            return tuple(zip(*results)) + (results[-1][1],)    # type: ignore

        return tuple(zip(*results)) + (None,)    # type: ignore

    def update_data(self, ax: Axes) -> None:
        """Update the plot after the axes have changed

        This is called after the user has panned or zoomed, and refreshes the
        plot.

        Args:
            ax: The Axes object we are drawing in
        """
        xmin, xmax = ax.viewLim.intervalx

        for event_type in _EVENT_TYPES:
            instances, start_times, durations, cutoff = self.get_data(
                    event_type, xmin, xmax)
            if instances:
                # update existing rectangles
                bars = self._bars[event_type].patches
                n_cur = len(instances)
                n_avail = len(bars)

                for i in range(min(n_cur, n_avail)):
                    bars[i].set_y(instances[i] - _BAR_WIDTH * 0.5)
                    bars[i].set_x(start_times[i])
                    bars[i].set_width(durations[i])
                    bars[i].set_visible(True)

                # set any superfluous ones invisible
                for i in range(n_cur, n_avail):
                    bars[i].set_visible(False)

            # update cutoff bars, if any
            if '_CUTOFF' in self._bars:
                bars = self._bars['_CUTOFF'].patches
                if cutoff:
                    for bar in bars:
                        bar.set_x(cutoff)
                        bar.set_width(self._global_xmax - cutoff)
                        bar.set_visible(True)
                    self._cutoff_warning.set_visible(True)
                else:
                    for bar in bars:
                        bar.set_visible(False)
                    self._cutoff_warning.set_visible(False)


tplot = None    # type: Optional[TimelinePlot]


def plot_timeline(performance_file: Path) -> None:
    global tplot
    tplot = TimelinePlot(performance_file)


def show_plots() -> None:
    """Actually show the plots on screen"""
    plt.show()      # type: ignore
    if tplot:
        tplot.close()



import logging
from pathlib import Path
from queue import Queue
from sqlite3 import Cursor
from threading import Thread
from typing import cast, Dict, Iterable, List, Optional, Tuple

from libmuscle.planner.planner import ResourceAssignment
from libmuscle.profiling import ProfileEvent, ProfileEventType
from libmuscle.manager.profile_database import ProfileDatabase
from ymmsl import Operator, Reference


_logger = logging.getLogger(__name__)


# When True causes add_events to return after the data is written.
# Used for testing only.
_SYNCHED = False


class ProfileStore(ProfileDatabase):
    """Creates and fills a profiling database.

    This class creates the database, and then allows writing to it.
    It's only used internally, and it's almost a non-const version
    of ProfileDatabase.
    """
    def __init__(self, db_dir: Path) -> None:
        """Create a new profile database.

        Args:
            db_file: The file to create and initialise.
        """
        db_file = db_dir / 'performance.sqlite'

        if db_file.exists():
            _logger.info(f'Overwriting profiling database {db_file}')
            try:
                # from Python 3.8, we can use missing_ok=True
                db_file.unlink()
            except FileNotFoundError:
                pass

        super().__init__(db_file)
        self._init_database()

        # 500 batches is about 250MB
        Item = Optional[Tuple[Reference, Iterable[ProfileEvent]]]
        self._queue: Queue[Item] = Queue(500)
        self._confirmation_queue: Queue[None] = Queue()
        self._thread = Thread(target=self._storage_thread, daemon=True)
        self._thread.start()

    def shutdown(self) -> None:
        """Shut down the profile store.

        This closes the database connection cleanly.
        """
        self._queue.put(None)
        self._thread.join()
        super().close()

    def store_instances(
            self, instances: List[Reference]) -> None:
        """Store names of instances in the simulation.

        Args:
            instances: List of instance names to store
        """
        cur = self._get_cursor()
        cur.execute("BEGIN IMMEDIATE TRANSACTION")

        cur.executemany(
                "INSERT INTO instances (name) VALUES (?)",
                [(str(name),) for name in instances])
        cur.execute("COMMIT")
        cur.close()

    def store_resources(self, resources: Dict[Reference, ResourceAssignment]) -> None:
        """Store resource assignments into the database.

        Args:
            resources: The resources to store.
        """
        cur = self._get_cursor()
        cur.execute("BEGIN IMMEDIATE TRANSACTION")

        for instance_id, res in resources.items():
            instance_oid = self._get_instance_oid(cur, instance_id)

            tuples = [
                    (instance_oid, node.node_name, core.cid)
                    for node in res.as_resources()
                    for core in node.cpu_cores]

            cur.executemany(
                    "INSERT INTO assigned_cores (instance_oid, node, core)"
                    " VALUES (?, ?, ?)",
                    tuples)

        cur.execute("COMMIT")
        cur.close()

    def add_events(
            self, instance_id: Reference, events: Iterable[ProfileEvent]
            ) -> None:
        """Adds profiling events to the database.

        Args:
            events: The events to add.
        """
        self._queue.put((instance_id, events))
        if _SYNCHED:
            self._confirmation_queue.get()

    def _storage_thread(self) -> None:
        """Background thread that stores the data.

        We're getting issues with the database being locked occasionally
        on slow file systems when we try to access it from multiple
        threads. So we use a single background thread now to do the
        writing.
        """
        Record = Tuple[
                int, int, float, float, Optional[str], Optional[int],
                Optional[int], Optional[int], Optional[int], Optional[int],
                Optional[float]]

        def to_tuple(e: ProfileEvent) -> Record:
            # Tell mypy this shouldn't happen
            assert e.start_time is not None
            assert e.stop_time is not None

            port_name = None if e.port is None else str(e.port.name)
            port_operator = None if e.port is None else e.port.operator.value

            return (
                    instance_oid, e.event_type.value, e.start_time.nanoseconds,
                    e.stop_time.nanoseconds, port_name, port_operator,
                    e.port_length, e.slot, e.message_number, e.message_size,
                    e.message_timestamp)

        cur = self._get_cursor()
        batch = self._queue.get()
        while batch is not None:
            instance_id, events = batch

            cur = self._get_cursor()
            cur.execute("BEGIN IMMEDIATE TRANSACTION")

            instance_oid = self._get_instance_oid(cur, instance_id)

            cur.executemany(
                    "INSERT INTO events"
                    " (instance_oid, event_type_oid, start_time, stop_time,"
                    "  port_name, port_operator_oid, port_length, slot,"
                    "  message_number, message_size, message_timestamp)"
                    " VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                    map(to_tuple, events))
            cur.execute("COMMIT")
            if _SYNCHED:
                self._confirmation_queue.put(None)
            batch = self._queue.get()

        cur.close()
        super().close()

    def _get_instance_oid(self, cur: Cursor, instance_id: Reference) -> int:
        """Get the oid for a given instance.

        Args:
            instance Instance id to look up

        Return:
            The oid used for it in the database
        """
        cur.execute(
                "SELECT oid FROM instances WHERE name = ?",
                (str(instance_id),))
        oids = cur.fetchall()
        return cast(int, oids[0][0])

    def _init_database(self) -> None:
        """Initialises the database.

        This creates the SQL tables needed to store the data.
        """
        cur = self._get_cursor()
        cur.execute("BEGIN IMMEDIATE TRANSACTION")
        cur.execute(
                "CREATE TABLE muscle3_format ("
                "    major_version INTEGER NOT NULL,"
                "    minor_version INTEGER NOT NULL)")
        cur.execute(
                "INSERT INTO muscle3_format(major_version, minor_version)"
                "    VALUES (1, 1)")

        cur.execute(
                "CREATE TABLE instances ("
                "    oid INTEGER PRIMARY KEY,"
                "    name TEXT UNIQUE)")

        cur.execute(
                "CREATE TABLE assigned_cores ("
                "    instance_oid INTEGER NOT NULL REFERENCES instances(oid),"
                "    node TEXT NOT NULL,"
                "    core INTEGER NOT NULL)")

        cur.execute(
                "CREATE TABLE event_types ("
                "    oid INTEGER PRIMARY KEY,"
                "    name TEXT UNIQUE)")
        event_types = [(t.value, t.name) for t in ProfileEventType]
        cur.executemany(
                "INSERT INTO event_types (oid, name) VALUES (?, ?)",
                event_types)

        cur.execute(
                "CREATE TABLE port_operators ("
                "    oid INTEGER PRIMARY KEY,"
                "    name TEXT UNIQUE)")
        port_operators = [(o.value, o.name) for o in Operator]
        cur.executemany(
                "INSERT INTO port_operators (oid, name) VALUES (?, ?)",
                port_operators)

        cur.execute(
                "CREATE TABLE events ("
                "    instance_oid INTEGER NOT NULL REFERENCES instances(oid),"
                "    event_type_oid INTEGER NOT NULL REFERENCES event_types(oid),"
                "    start_time INTEGER NOT NULL,"
                "    stop_time INTEGER NOT NULL,"
                "    port_name TEXT,"
                "    port_operator_oid INTEGER REFERENCES port_operators(oid),"
                "    port_length INTEGER,"
                "    slot INTEGER,"
                "    message_number INTEGER,"
                "    message_size INTEGER,"
                "    message_timestamp DOUBLE)")

        cur.execute("CREATE INDEX instances_oid_idx ON instances(oid)")

        cur.execute("CREATE INDEX events_start_time_idx ON events(start_time)")

        cur.execute(
                "CREATE VIEW all_events"
                " AS SELECT"
                "    i.name AS instance, et.name AS type,"
                "    e.start_time AS start_time, e.stop_time AS stop_time,"
                "    e.port_name AS port, o.name AS operator,"
                "    e.port_length AS port_length, e.slot AS slot,"
                "    e.message_number AS message_number,"
                "    e.message_size AS message_size,"
                "    e.message_timestamp AS message_timestamp"
                " FROM"
                "    events e"
                "    JOIN instances i ON e.instance_oid = i.oid"
                "    LEFT JOIN event_types et ON e.event_type_oid = et.oid"
                "    LEFT JOIN port_operators o ON e.port_operator_oid = o.oid")

        cur.execute("COMMIT")
        cur.close()



import itertools
from pathlib import Path
import sys
from typing import Generator, List, Optional, cast

from ymmsl import Reference


def instance_to_kernel(instance: Reference) -> Reference:
    """Extracts the name of the kernel from an instance name.

    Args:
        instance: The name of a kernel instance.

    Returns:
        The name of its kernel.
    """
    i = len(instance)
    while isinstance(instance[i - 1], int):
        i -= 1
    return instance[:i]


def instance_indices(instance: Reference) -> List[int]:
    """Extracts the indices from an instance name.

    Args:
        instance: The name of a kernel instance.

    Returns:
        The name of its kernel.
    """
    i = len(instance)
    while isinstance(instance[i - 1], int):
        i -= 1

    # Note that the slice operator on References returns a Reference,
    # which we don't want here.
    return [cast(int, instance[j]) for j in range(i, len(instance))]


def generate_indices(dims: List[int]) -> Generator[List[int], None, None]:
    """Generates all indices in a block of the given dimensions.

    Args:
        dims: The dimensions of the block.

    Yields:
        Lists of indices, one for each point in the block.
    """
    for index in itertools.product(*map(range, dims)):
        yield list(index)


def extract_log_file_location(filename: str) -> Optional[Path]:
    """Gets the log file location from the command line.

    Extracts the --muscle-log-file=<path> argument to tell the
    MUSCLE library where to write the local log file. This
    function will extract this argument from the command line
    arguments if it is present. If the given path is to a
    directory, <filename> will be written inside of that directory,
    if the path is not an existing directory, then it will be used
    as the name of the log file to write to. If no command line
    argument is given, this function returns None.

    Args:
        filename: Default file name to use.

    Returns:
        Path to the log file to write.
    """
    # Neither getopt, optparse, or argparse will let me pick out
    # just one option from the command line and ignore the rest.
    # So we do it by hand.
    prefix = '--muscle-log-file='
    given_path_str = None
    for arg in sys.argv[1:]:
        if arg.startswith(prefix):
            given_path_str = arg[len(prefix):]

    if not given_path_str:
        return None

    given_path = Path(given_path_str)

    if given_path.is_dir():
        return given_path / filename
    return given_path



import pytest
from ymmsl import Settings

from libmuscle.communicator import Message
from libmuscle.snapshot import Snapshot, MsgPackSnapshot, SnapshotMetadata


@pytest.fixture
def snapshot() -> Snapshot:
    triggers = ['test triggers']
    wallclock_time = 15.3
    port_message_counts = {'in': [1], 'out': [4], 'muscle_settings_in': [0]}
    is_final = True
    message = Message(1.2, data='test_data')
    snapshot = MsgPackSnapshot(
            triggers, wallclock_time, port_message_counts, is_final, message,
            Settings({'test': 1}))
    assert snapshot.triggers == triggers
    assert snapshot.wallclock_time == wallclock_time
    assert snapshot.port_message_counts == port_message_counts
    assert snapshot.is_final_snapshot == is_final
    assert snapshot.message == message
    assert snapshot.settings_overlay.keys() == {'test'}
    assert snapshot.settings_overlay['test'] == 1
    return snapshot


def test_snapshot(snapshot: Snapshot) -> None:
    assert isinstance(snapshot, Snapshot)

    binary_snapshot = snapshot.to_bytes()
    assert isinstance(binary_snapshot, bytes)

    snapshot2 = MsgPackSnapshot.from_bytes(binary_snapshot)

    assert snapshot2.triggers == snapshot.triggers
    assert snapshot2.wallclock_time == snapshot.wallclock_time
    assert snapshot2.port_message_counts == snapshot.port_message_counts
    assert snapshot2.is_final_snapshot == snapshot.is_final_snapshot
    assert snapshot2.message.timestamp == snapshot.message.timestamp
    assert snapshot2.message.next_timestamp == snapshot.message.next_timestamp
    assert snapshot2.message.data == snapshot.message.data


def test_snapshot_metadata(snapshot: Snapshot) -> None:
    metadata = SnapshotMetadata.from_snapshot(snapshot, 'test')

    assert metadata.triggers == snapshot.triggers
    assert metadata.wallclock_time == snapshot.wallclock_time
    assert metadata.port_message_counts == snapshot.port_message_counts
    assert metadata.is_final_snapshot == snapshot.is_final_snapshot
    assert metadata.timestamp == snapshot.message.timestamp
    assert metadata.next_timestamp == snapshot.message.next_timestamp
    assert metadata.snapshot_filename == 'test'


def test_message_with_settings() -> None:
    message = Message(1.0, 2.0, 'test_data', Settings({'setting': True}))
    snapshot = MsgPackSnapshot([], 0, {}, False, message, Settings())
    assert snapshot.message.settings.get('setting') is True

    binary_snapshot = snapshot.to_bytes()
    assert isinstance(binary_snapshot, bytes)

    snapshot2 = MsgPackSnapshot.from_bytes(binary_snapshot)
    assert snapshot2.message.settings.get('setting') is True


def test_implicit_snapshot() -> None:
    message = None
    snapshot = MsgPackSnapshot([], 0, {}, True, message, Settings())
    assert snapshot.message is None

    binary_snapshot = snapshot.to_bytes()
    assert isinstance(binary_snapshot, bytes)

    snapshot2 = MsgPackSnapshot.from_bytes(binary_snapshot)
    assert snapshot2.message is None



from errno import ENOTCONN
import select
import logging
import socket
from typing import Optional, Tuple

from libmuscle.mcp.transport_client import ProfileData, TransportClient, TimeoutHandler
from libmuscle.mcp.tcp_util import recv_all, recv_int64, send_int64
from libmuscle.profiling import ProfileTimestamp


_logger = logging.getLogger(__name__)


class TcpTransportClient(TransportClient):
    """A client that connects to a TCPTransport server.
    """
    @staticmethod
    def can_connect_to(location: str) -> bool:
        """Whether this client class can connect to the given location.

        Args:
            location: The location to potentially connect to.

        Returns:
            True iff this class can connect to this location.
        """
        return location.startswith('tcp:')

    def __init__(self, location: str) -> None:
        """Create a TcpClient for a given location.

        The client will connect to this location and be able to request
        messages from any instance and port represented by it.

        Args:
            location: A location string for the peer.
        """
        addresses = location[4:].split(',')

        sock: Optional[socket.SocketType] = None
        for address in addresses:
            try:
                sock = self._connect(address, False)
                break
            except RuntimeError:
                pass

        if sock is None:
            # None of our quick connection attempts worked. Either there's a network
            # problem, or the server is very busy. Let's try again with more patience.
            _logger.warning(
                    f'Could not immediately connect to {location}, trying again with'
                    ' more patience. Please report this if it happens frequently.')

            for address in addresses:
                try:
                    sock = self._connect(address, True)
                    break
                except RuntimeError:
                    pass

        if sock is None:
            _logger.error(f'Failed to connect also on the second try to {location}')
            raise RuntimeError('Could not connect to the server at location'
                               ' {}'.format(location))

        if hasattr(socket, "TCP_NODELAY"):
            sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
        if hasattr(socket, "TCP_QUICKACK"):
            sock.setsockopt(socket.SOL_TCP, socket.TCP_QUICKACK, 1)
        self._socket = sock

        if hasattr(select, "poll"):
            self._poll_obj: Optional[select.poll] = select.poll()
            self._poll_obj.register(self._socket, select.POLLIN)
        else:
            self._poll_obj = None  # On platforms that don't support select.poll

    def call(self, request: bytes, timeout_handler: Optional[TimeoutHandler] = None
             ) -> Tuple[bytes, ProfileData]:
        """Send a request to the server and receive the response.

        This is a blocking call.

        Args:
            request: The request to send
            timeout_handler: Optional timeout handler. This is used for communication
                deadlock detection.

        Returns:
            The received response
        """
        start_wait = ProfileTimestamp()
        send_int64(self._socket, len(request))
        self._socket.sendall(request)

        did_timeout = False
        if timeout_handler is not None:
            while not self._poll(timeout_handler.timeout):
                did_timeout = True
                timeout_handler.on_timeout()

        length = recv_int64(self._socket)
        if did_timeout:
            assert timeout_handler is not None  # mypy
            timeout_handler.on_receive()
        start_transfer = ProfileTimestamp()

        response = recv_all(self._socket, length)
        stop_transfer = ProfileTimestamp()
        return response, (start_wait, start_transfer, stop_transfer)

    def _poll(self, timeout: float) -> bool:
        """Poll the socket and return whether its ready for receiving.

        This method blocks until the socket is ready for receiving, or :param:`timeout`
        seconds have passed (whichever is earlier).

        Args:
            timeout: timeout in seconds

        Returns:
            True if the socket is ready for receiving data, False otherwise.
        """
        if self._poll_obj is not None:
            ready = self._poll_obj.poll(timeout * 1000)  # poll timeout is in ms
        else:
            # Fallback to select()
            ready, _, _ = select.select([self._socket], (), (), timeout)
        return bool(ready)

    def close(self) -> None:
        """Closes this client.

        This closes any connections this client has and/or performs
        other shutdown activities.
        """
        try:
            self._socket.shutdown(socket.SHUT_RDWR)
            self._socket.close()
        except OSError as e:
            # This can happen if the peer has shut down already when we
            # close our connection to it, which is fine.
            if e.errno != ENOTCONN:
                raise

    def _connect(self, address: str, patient: bool) -> socket.SocketType:
        loc_parts = address.rsplit(':', 1)
        host = loc_parts[0]
        if host.startswith('['):
            if host.endswith(']'):
                host = host[1:-1]
            else:
                raise RuntimeError('Invalid address')
        port = int(loc_parts[1])

        addrinfo = socket.getaddrinfo(
                host, port, 0, socket.SOCK_STREAM, socket.IPPROTO_TCP)

        for family, socktype, proto, _, sockaddr in addrinfo:
            try:
                sock = socket.socket(family, socktype, proto)
            except Exception:
                continue

            try:
                sock.settimeout(20.0 if patient else 3.0)     # seconds
                sock.connect(sockaddr)
                sock.settimeout(60.0)
            except Exception:
                sock.close()
                continue
            return sock

        raise RuntimeError('Could not connect')



from typing import List

import pytest
from ymmsl import Reference

from libmuscle.util import instance_indices, instance_to_kernel


@pytest.fixture
def instances() -> List[Reference]:
    return [
        Reference('test'),
        Reference('test.test'),
        Reference('test[4]'),
        Reference('test[3][2]'),
        Reference('test.test[3]'),
        Reference('test.test[1][2]')]


def test_instance_to_kernel(instances: List[Reference]) -> None:
    assert instance_to_kernel(instances[0]) == 'test'
    assert instance_to_kernel(instances[1]) == 'test.test'
    assert instance_to_kernel(instances[2]) == 'test'
    assert instance_to_kernel(instances[3]) == 'test'
    assert instance_to_kernel(instances[4]) == 'test.test'
    assert instance_to_kernel(instances[5]) == 'test.test'


def test_instance_indices(instances: List[Reference]) -> None:
    assert instance_indices(instances[0]) == []
    assert instance_indices(instances[1]) == []
    assert instance_indices(instances[2]) == [4]
    assert instance_indices(instances[3]) == [3, 2]
    assert instance_indices(instances[4]) == [3]
    assert instance_indices(instances[5]) == [1, 2]



from enum import Enum
from time import perf_counter_ns, time_ns
from typing import Optional

from ymmsl import Port


class ProfileEventType(Enum):
    """Profiling event types for MUSCLE3."""
    REGISTER = 0
    CONNECT = 4
    SEND = 2
    RECEIVE = 3
    RECEIVE_WAIT = 5
    RECEIVE_TRANSFER = 6
    RECEIVE_DECODE = 7
    SHUTDOWN_WAIT = 9
    DISCONNECT_WAIT = 8
    SHUTDOWN = 10
    DEREGISTER = 1


class ProfileTimestamp:
    """A timestamp for profiling.

    This has higher resolution than Timestamp, storing a number of
    nanoseconds since the UNIX epoch in an int.

    Attributes:
        nanoseconds: Nanoseconds since the UNIX epoch.
    """
    _time_ref = time_ns() - perf_counter_ns()

    def __init__(self, nanoseconds: Optional[int] = None) -> None:
        """Create a timestamp representing now.

        Args:
            nanoseconds: Time to set. If None, use the current time.
        """
        if nanoseconds is None:
            nanoseconds = perf_counter_ns() + self._time_ref
        self.nanoseconds = nanoseconds


class ProfileEvent:
    """A profile event as used by MUSCLE3.

    This represents a single measurement of the timing of some event
    that occurred while executing the simulation.

    Args:
        event_type: Type of event that was measured.
        start_time: When the event started (real-world, not
                simulation time).
        stop_time: When the event ended (real-world, not simulation
                time).
        port: Port used for sending or receiving, if applicable.
        port_length: Length of that port, if a vector.
        slot: Slot that was sent or received on, if applicable.
        message_number: Number of message on this port, if applicable.
                Starts at 0 for the first message sent or received.
        message_size: Size of the message involved, if applicable.
        message_timestamp: Timestamp sent with the message, if
                applicable.

    Attributes:
        event_type: Type of event that was measured.
        start_time: When the event started (real-world, not
                simulation time).
        stop_time: When the event ended (real-world, not simulation
                time).
        port: Port used for sending or receiving, if applicable.
        port_length: Length of that port, if a vector.
        slot: Slot that was sent or received on, if applicable.
        message_number: Number of message on this port, if applicable.
                Starts at 0 for the first message sent or received.
        message_size: Size of the message involved, if applicable.
        message_timestamp: Timestamp sent with the message, if
                applicable.
    """
    def __init__(
            self,
            event_type: ProfileEventType,
            start_time: Optional[ProfileTimestamp] = None,
            stop_time: Optional[ProfileTimestamp] = None,
            port: Optional[Port] = None,
            port_length: Optional[int] = None,
            slot: Optional[int] = None,
            message_number: Optional[int] = None,
            message_size: Optional[int] = None,
            message_timestamp: Optional[float] = None
            ) -> None:

        self.event_type = event_type
        self.start_time = start_time
        self.stop_time = stop_time
        self.port = port
        self.port_length = port_length
        self.slot = slot
        self.message_number = message_number
        self.message_size = message_size
        self.message_timestamp = message_timestamp

    def start(self) -> None:
        """Sets start_time to the current time.
        """
        self.start_time = ProfileTimestamp()

    def stop(self) -> None:
        """Sets stop_time to the current time.
        """
        self.stop_time = ProfileTimestamp()



from enum import Enum
import logging
from socket import gethostname
from typing import List, Optional

import psutil

from libmuscle.native_instantiator.slurm import slurm


_logger = logging.getLogger(__name__)


class Scheduler(Enum):
    NONE = 0
    SLURM = 1


class GlobalResources:
    """Detects available compute resources.

    This detects whether we're running locally or in a SLURM allocation, and returns
    available resources on request. This class describes all the available resources,
    not the ones local to a node.

    Attributes:
        scheduler: The HPC scheduler we're running under, if any.
        nodes: List of hostnames of available nodes to run on.
        logical_cpus_per_node: Number of cores available on each node.
                List alongside nodes.
    """
    def __init__(self) -> None:
        """Create a GlobalResources.

        Detects available resources and initialises the object, which can then be
        queried.
        """
        if slurm().in_slurm_allocation():
            _logger.info('Detected a SLURM allocation')
            self.scheduler = Scheduler.SLURM
            self.nodes = slurm().get_nodes()
            self.logical_cpus_per_node = slurm().get_logical_cpus_per_node()
            _logger.info(
                    f'We have {len(self.nodes)} nodes and a total of'
                    f' {sum(self.logical_cpus_per_node)} logical CPUs available')
        else:
            _logger.info('Running locally without a cluster scheduler')
            self.scheduler = Scheduler.NONE
            self.nodes = [gethostname()]
            self.logical_cpus_per_node = [psutil.cpu_count(logical=True) or 0]
            _logger.info(
                    f'We have {self.logical_cpus_per_node[0]} logical CPUS available')

    def on_cluster(self) -> bool:
        """Return whether we're running on a cluster."""
        return self.scheduler != Scheduler.NONE

    def agent_launch_command(self, agent_cmd: List[str]) -> List[str]:
        """Return a command for launching one agent on each node.

        Args:
            agent_cmd: A command that will start the agent.
        """
        if self.scheduler == Scheduler.SLURM:
            return slurm().agent_launch_command(agent_cmd, len(self.nodes))
        return agent_cmd


_global_resources: Optional[GlobalResources] = None
"""Global resources object.

This is a singleton, and that's fine because it's created once and then read-only. Also,
it's used in two places, and making two objects logs everything twice which is annoying.
"""


def global_resources() -> GlobalResources:
    """Wrapper for _global_resources.

    This is here to ensure that the object gets created after we've configured logging,
    so that the log output it generates actually ends up in the manager log.

    The users are all in the main thread of the NativeInstantiator background process,
    so there's no need for a lock right now.
    """
    global _global_resources

    if _global_resources is None:
        _global_resources = GlobalResources()

    return _global_resources



import pathlib
from setuptools import find_packages, setup


_here = pathlib.Path(__file__).resolve().parent
_version_file = _here / 'VERSION'
with _version_file.open('r') as f:
    _version = f.read().strip()

# this is a hack, but it works for now and keeps the version number in a single file
_libmuscle_version_file = _here / 'libmuscle' / 'python' / 'libmuscle' / 'version.py'
with _libmuscle_version_file.open('w') as f:
    f.write('__version__ = \'{}\'\n'.format(_version))

_muscle3_packages = [
        p for p in find_packages() + find_packages('libmuscle/python')
        if p != 'integration_test']

_long_desc = (_here / 'README.rst').read_text()

setup(
    name='muscle3',
    version=_version,
    description='Version 3 of the MUltiScale Coupling Library and Environment',
    long_description=_long_desc,
    long_description_content_type='text/x-rst',
    author='Lourens Veen',
    author_email='l.veen@esciencecenter.nl',
    url='https://github.com/multiscale/muscle3',
    license='Apache License 2.0',
    keywords=['multiscale', 'coupling', 'MUSCLE'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        ],

    packages=_muscle3_packages,
    package_dir={
        'muscle3': 'muscle3',
        'libmuscle': 'libmuscle/python/libmuscle'
    },
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'muscle_manager=muscle3.muscle_manager:manage_simulation',
            'muscle3=muscle3.muscle3:muscle3']
    },
    python_requires='>=3.8, <4',
    install_requires=[
        'click>=7.1,<9',
        'matplotlib>=3,<4',
        'msgpack>=1,<2',
        'psutil>=5.0.0',
        'parsimonious',
        "numpy>=1.22",
        'ymmsl>=0.14.0,<0.15'          # Also in CI, and examples requirements.txt
    ],
    extras_require={
        'dev': [
            'tox'
        ]
    },
)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, List, Optional, cast

import msgpack
from ymmsl import Reference, Settings

from libmuscle.mpp_message import MPPMessage
from libmuscle import communicator


class Snapshot(ABC):
    """Snapshot data structure.

    This is an abstract base class, implementations are provided by subclasses.
    """
    SNAPSHOT_VERSION_BYTE = b'\0'

    def __init__(self,
                 triggers: List[str],
                 wallclock_time: float,
                 port_message_counts: Dict[str, List[int]],
                 is_final_snapshot: bool,
                 message: Optional['communicator.Message'],
                 settings_overlay: Settings) -> None:
        self.triggers = triggers
        self.wallclock_time = wallclock_time
        self.port_message_counts = port_message_counts
        self.is_final_snapshot = is_final_snapshot
        self.message = message
        # self.message is None for implicit snapshots, so we cannot store the
        # Settings overlay in that message object.
        self.settings_overlay = settings_overlay

    @classmethod
    @abstractmethod
    def from_bytes(cls, data: bytes) -> 'Snapshot':
        """Create a snapshot object from binary data.

        Args:
            data: binary data representing the snapshot. Note that this must
                **exclude** the versioning byte.
        """
        ...

    @abstractmethod
    def to_bytes(self) -> bytes:
        """Convert the snapshot object to binary data.

        Returns:
            Binary data representing the snapshot. Note that this must
                **exclude** the versioning byte.
        """
        ...


class MsgPackSnapshot(Snapshot):
    """Snapshot stored in messagepack format
    """
    SNAPSHOT_VERSION_BYTE = b'1'

    @classmethod
    def from_bytes(cls, data: bytes) -> 'Snapshot':
        dct = msgpack.loads(data)
        return cls(dct['triggers'],
                   dct['wallclock_time'],
                   dct['port_message_counts'],
                   dct['is_final_snapshot'],
                   cls.bytes_to_message(dct['message']),
                   Settings(dct['settings_overlay']))

    def to_bytes(self) -> bytes:
        return cast(bytes, msgpack.dumps({
            'triggers': self.triggers,
            'wallclock_time': self.wallclock_time,
            'port_message_counts': self.port_message_counts,
            'is_final_snapshot': self.is_final_snapshot,
            'message': self.message_to_bytes(self.message),
            'settings_overlay': self.settings_overlay.as_ordered_dict()
        }))

    @staticmethod
    def message_to_bytes(message: Optional['communicator.Message']) -> bytes:
        """Use MPPMessage serializer for serializing the message object
        """
        if message is None:
            return b''
        settings = Settings()
        if message.settings is not None:
            settings = message.settings
        return MPPMessage(Reference('_'), Reference('_'), None,
                          message.timestamp, message.next_timestamp,
                          settings, 0, -1.0, message.data).encoded()

    @staticmethod
    def bytes_to_message(data: bytes) -> Optional['communicator.Message']:
        """Use MPPMessage deserializer for serializing the message object
        """
        if not data:
            return None
        mpp_message = MPPMessage.from_bytes(data)
        return communicator.Message(mpp_message.timestamp,
                                    mpp_message.next_timestamp,
                                    mpp_message.data,
                                    mpp_message.settings_overlay)


@dataclass
class SnapshotMetadata:
    """Metadata of a snapshot for sending to the muscle_manager.
    """
    triggers: List[str]
    wallclock_time: float
    timestamp: float
    next_timestamp: Optional[float]
    port_message_counts: Dict[str, List[int]]
    is_final_snapshot: bool
    # storing as str, because Path cannot be serialized by msgpack
    snapshot_filename: str

    @staticmethod
    def from_snapshot(snapshot: Snapshot, snapshot_filename: str
                      ) -> 'SnapshotMetadata':
        """Create snapshot metadata from the given snapshot and filename
        """
        return SnapshotMetadata(
            snapshot.triggers,
            snapshot.wallclock_time,
            snapshot.message.timestamp if snapshot.message else float('NaN'),
            snapshot.message.next_timestamp if snapshot.message else None,
            snapshot.port_message_counts,
            snapshot.is_final_snapshot,
            snapshot_filename
        )



import logging
from pathlib import Path
from subprocess import Popen
from typing import Dict, List, Tuple


_logger = logging.getLogger(__name__)


class ProcessManager:
    """Manages a set of running processes."""
    def __init__(self) -> None:
        """Create a ProcessManager."""
        self._processes: Dict[str, Popen] = dict()

    def start(
            self, name: str, work_dir: Path, args: List[str], env: Dict[str, str],
            stdout: Path, stderr: Path) -> None:
        """Start a process.

        The files that the output is directed to will be overwritten if they already
        exist.

        Args:
            name: Name under which this process will be known
            work_dir: Working directory in which to start
            args: Executable and arguments to run
            env: Environment variables to set
            stdout: File to redirect stdout to
            stderr: File to redirect stderr to

        Raises:
            RuntimeError: If there is already a process with the given name.
            OSError: If the process could not be started.
        """
        if name in self._processes:
            raise RuntimeError(f'Process {name} already exists')
        _logger.debug(f'Starting process {args} with env {env} in {work_dir}')
        with stdout.open('w') as out, stderr.open('w') as err:
            self._processes[name] = Popen(
                    args, cwd=work_dir, env=env, stdout=out, stderr=err)

    def cancel_all(self) -> None:
        """Stops all running processes.

        This does not wait for them to terminate, it just sends the signal to kill
        them.
        """
        for process in self._processes.values():
            process.kill()

    def get_finished(self) -> List[Tuple[str, int]]:
        """Returns names and exit codes of finished processes.

        This returns all processes that have finished running since the previous call;
        each started process will be returned exactly once.
        """
        result: List[Tuple[str, int]] = list()

        for name, process in self._processes.items():
            exit_code = process.poll()
            if exit_code is not None:
                result.append((name, exit_code))

        for name, _ in result:
            del self._processes[name]

        return result
